'''
This is the default encoder used by clients for all messages sent with format 
defined as "json".  This default can be changed by reconfiguring your client:
  
  deviceCli.setMessageEncoderModule("json", myCustomEncoderModule)

'''
import json
from datetime import datetime
import pytz
from ibmiotf import Message, InvalidEventException

'''
Convert Python dictionary object into a UTF-8 encoded JSON string.  Timestamp information is
not passed into the encoded message.
'''
def encode(data=None, timestamp=None):
    return json.dumps(data)

'''
Convert a generic JSON message

* The entire message is converted to JSON and treated as the message data
* The timestamp of the message is the time that the message is RECEIVED
'''
def decode(message):
    try:
        data = json.loads(message.payload.decode("utf-8"))
    except ValueError as e:
        raise InvalidEventException("Unable to parse JSON.  payload=\"%s\" error=%s" % (message.payload, str(e)))
    
    timestamp = datetime.now(pytz.timezone('UTC'))
    
    # TODO: Flatten JSON, covert into array of key/value pairs
    return Message(data, timestamp)



'''
This is the default encoder used by clients for all messages sent with format
defined as "xml".  This default can be changed by reconfiguring your client:

  deviceCli.setMessageEncoderModule("xml", myCustomEncoderModule)

'''
import json
import dicttoxml
import xmltodict
from datetime import datetime
import pytz
from ibmiotf import Message, InvalidEventException

'''
Convert Python dictionary object into a UTF-8 encoded XML string.  Timestamp information is
not passed into the encoded message.
'''
def encode(data=None, timestamp=None):
    return dicttoxml.dicttoxml(data)

'''
Convert XML data into Python Dictionary Object
The entire message is converted to JSON and treated as the message data
The timestamp of the message is the time that the message is RECEIVED
'''
def decode(message):
    try:
        jsonData = json.dumps(xmltodict.parse(message.payload))
        data = json.loads(jsonData.decode("utf-8"))
    except ValueError as e:
        raise InvalidEventException("Unable to parse XML.  payload=\"%s\" error=%s" % (message.payload, str(e)))
    timestamp = datetime.now(pytz.timezone('UTC'))
    return Message(data, timestamp)



# *****************************************************************************
# Copyright (c) 2016 IBM Corporation and other Contributors.
#
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#
# Contributors:
#   Lokesh Haralakatta  - Initial Contribution
# *****************************************************************************

import ibmiotf.device
import ibmiotf.application
import uuid
import os
from ibmiotf import *
from nose.tools import *
from nose import SkipTest
import logging
import testUtils

class TestDevice(testUtils.AbstractTest):
    registeredDevice = None
    managedClient = None

    DEVICE_TYPE = "test_device"
    DEVICE_ID = str(uuid.uuid4())

    @classmethod
    def setup_class(self):
        try:
            deviceType = self.setupAppClient.api.getDeviceType(self.DEVICE_TYPE)
        except APIException as e:
            if e.httpCode == 404:
                deviceType = self.setupAppClient.api.addDeviceType(self.DEVICE_TYPE)
            else:
                raise e

        self.registeredDevice = self.setupAppClient.api.registerDevice(self.DEVICE_TYPE, self.DEVICE_ID)

        self.options={
            "org": self.ORG_ID,
            "type": self.registeredDevice["typeId"],
            "id": self.registeredDevice["deviceId"],
            "auth-method": "token",
            "auth-token": self.registeredDevice["authToken"]
        }

        #Create default DeviceInfo Instance and associate with ManagedClient Instance
        deviceInfoObj = ibmiotf.device.DeviceInfo()
        deviceInfoObj.fwVersion = 0.0
        self.managedClient = ibmiotf.device.ManagedClient(self.options, deviceInfo=deviceInfoObj)

    @classmethod
    def teardown_class(self):
        del self.managedClient
        self.setupAppClient.api.deleteDevice(self.DEVICE_TYPE, self.DEVICE_ID)


    @raises(Exception)
    def testManagedClientQSException(self):
        with assert_raises(Exception)as e:
            options={"org": "quickstart", "type": self.registeredDevice["typeId"], "id": self.registeredDevice["deviceId"],
                                        "auth-method":"None", "auth-token":"None" }
            ibmiotf.device.ManagedClient(options)
        assert_equals(e.exception, Exception)

    @SkipTest
    def testManagedClientSetMethods(self):
        self.managedClient.connect()
        #Define device properties to be notified whenever reset
        self.managedClient._deviceMgmtObservations = ["deviceInfo.manufacturer", "deviceInfo.descriptiveLocation",
                                                "deviceInfo.fwVersion", "deviceInfo.model", "deviceInfo.description",
                                                "deviceInfo.deviceClass", "deviceInfo.hwVersion", "deviceInfo.serialNumber"]

        #Reset managedClient properties  and validate the returned is instance of threading.Event
        if sys.version_info[0] < 3:
            assert_is_instance(self.managedClient.setErrorCode(1),threading._Event)
            assert_is_instance(self.managedClient.setLocation(longitude=200, latitude=278),threading._Event)
            assert_is_instance(self.managedClient.setSerialNumber('iot-device-12345'),threading._Event)
            assert_is_instance(self.managedClient.setManufacturer("IBM India Pvt Ltd"),threading._Event)
            assert_is_instance(self.managedClient.setModel("2016"),threading._Event)
            assert_is_instance(self.managedClient.setdeviceClass("Smart Device"),threading._Event)
            assert_is_instance(self.managedClient.setDescription("Sample Smart IoT Device"),threading._Event)
            assert_is_instance(self.managedClient.setFwVersion("1.0"),threading._Event)
            assert_is_instance(self.managedClient.setHwVersion("2.0"),threading._Event)
            assert_is_instance(self.managedClient.setDescriptiveLocation("ISL Lab Bangalore"),threading._Event)
            assert_is_instance(self.managedClient.clearErrorCodes(),threading._Event)
            assert_is_instance(self.managedClient.addLog(),threading._Event)
            assert_is_instance(self.managedClient.clearLog(),threading._Event)
            assert_is_instance(self.managedClient.unmanage(),threading._Event)
        else:
            assert_is_instance(self.managedClient.setErrorCode(1),threading.Event)
            assert_is_instance(self.managedClient.setLocation(longitude=200, latitude=278),threading.Event)
            assert_is_instance(self.managedClient.setSerialNumber('iot-device-12345'),threading.Event)
            assert_is_instance(self.managedClient.setManufacturer("IBM India Pvt Ltd"),threading.Event)
            assert_is_instance(self.managedClient.setModel("2016"),threading.Event)
            assert_is_instance(self.managedClient.setdeviceClass("Smart Device"),threading.Event)
            assert_is_instance(self.managedClient.setDescription("Sample Smart IoT Device"),threading.Event)
            assert_is_instance(self.managedClient.setFwVersion("1.0"),threading.Event)
            assert_is_instance(self.managedClient.setHwVersion("2.0"),threading.Event)
            assert_is_instance(self.managedClient.setDescriptiveLocation("ISL Lab Bangalore"),threading.Event)
            assert_is_instance(self.managedClient.clearErrorCodes(),threading.Event)
            assert_is_instance(self.managedClient.addLog(),threading.Event)
            assert_is_instance(self.managedClient.clearLog(),threading.Event)
            assert_is_instance(self.managedClient.unmanage(),threading.Event)

        self.managedClient.disconnect()

    @SkipTest
    def testDeviceRebootAction(self):
        def rebootActionCB(reqId,action):
            print("Device rebootActionCB called")

        mgmtRequest = {"action": "device/reboot", "parameters": [{"name": "action","value": "reboot" }],
                       "devices": [{ "typeId": self.registeredDevice["typeId"], "deviceId": self.registeredDevice["deviceId"] }]}

        #Setup user defined reboot call back
        self.managedClient.deviceActionCallback = rebootActionCB
        self.managedClient.connect()


        #Initialize device management request to reboot
        initResult = self.apiClient.initiateDeviceManagementRequest(mgmtRequest)
        reqId = initResult['reqId']

        time.sleep(1)

        #Delete initiated device management request
        assert_true(self.apiClient.deleteDeviceManagementRequest(reqId))

        self.managedClient.disconnect()

    @SkipTest
    def testDeviceFactoryResetAction(self):
        def factoryResetActionCB(reqId,action):
            print("Device factoryResetActionCB called")

        mgmtRequest = {"action": "device/factoryReset", "parameters": [{"name": "action","value": "reset" }],
                       "devices": [{ "typeId": self.registeredDevice["typeId"], "deviceId": self.registeredDevice["deviceId"] }]}

        #Setup user defined factory reset call back
        self.managedClient.deviceActionCallback = factoryResetActionCB
        self.managedClient.connect()

        #Initialize device management request for factory reset
        initResult = self.apiClient.initiateDeviceManagementRequest(mgmtRequest)
        reqId = initResult['reqId']

        time.sleep(1)

        #Delete initiated device management request
        assert_true(self.apiClient.deleteDeviceManagementRequest(reqId))

        self.managedClient.disconnect()

    @SkipTest
    def testFirmwareDownloadAction(self):
        def downloadHandler(client,info):
            try:
                print("Setting ManagedClient.UPDATESTATE_DOWNLOADING")
                client.setState(device.ManagedClient.UPDATESTATE_DOWNLOADING)
                print("Setting ManagedClient.UPDATESTATE_DOWNLOADED")
                client.setState(device.ManagedClient.UPDATESTATE_DOWNLOADED)
            except Exception :
                print("Exception from downloadHandler")

        def firmwareDownloadActionCB(action,devInfo):
            print("Device firmwareDownloadActionCB called")
            self.managedClient.setState(device.ManagedClient.UPDATESTATE_IDLE)
            print("Calling downloadHandler Thread")

            dThread = threading.Thread(target=downloadHandler,args=(self.managedClient,devInfo))
            dThread.start()
            dThread.join()
            print("downloadHandler Thread is done")


        mgmtRequest = {"action": "firmware/download", "parameters": [{"name": "version", "value": "0.1.11" },
                       {"name": "name", "value": "RasPi01 firmware"}, {"name": "verifier", "value": "123df"},
                       {"name": "uri","value": "https://github.com/ibm-messaging/iot-raspberrypi/releases/download/1.0.2.1/iot_1.0-2_armhf.deb"}],
                       "devices": [{"typeId": self.registeredDevice["typeId"],"deviceId": self.registeredDevice["deviceId"]}]};

        self.managedClient.__firmwareUpdate = device.DeviceFirmware('0.0','0.0','uri','verifier',device.ManagedClient.UPDATESTATE_IDLE,
                                                      device.ManagedClient.UPDATESTATE_IDLE,'updatedDateTime')

        #Setup user defined firmware download call back
        self.managedClient.firmwereActionCallback = firmwareDownloadActionCB
        self.managedClient.connect()
        self.managedClient.setState(device.ManagedClient.UPDATESTATE_IDLE)
        self.managedClient.setUpdateStatus(device.ManagedClient.UPDATESTATE_IDLE)

        #Initialize device management request for firmware download
        initResult = self.apiClient.initiateDeviceManagementRequest(mgmtRequest)
        reqId = initResult['reqId']
        time.sleep(1)

        #Delete initiated device management request
        assert_true(self.apiClient.deleteDeviceManagementRequest(reqId))

        self.managedClient.disconnect()

    @SkipTest
    def testFirmwareUpdateAction(self):
        def updateHandler(client,info):
            try:
                print("Setting ManagedClient.UPDATESTATE_IN_PROGRESS")
                client.setUpdateStatus(device.ManagedClient.UPDATESTATE_IN_PROGRESS)
                print("Setting ManagedClient.UPDATESTATE_SUCCESS")
                threading.Timer(5,client.setUpdateStatus,[device.ManagedClient.UPDATESTATE_SUCCESS]).start()

            except Exception :
                print("Exception from updateHandler")

        def firmwareUpdateActionCB(action,devInfo):
            print("Device firmwareUpdateActionCB called")
            self.managedClient.setUpdateStatus(device.ManagedClient.UPDATESTATE_IDLE)
            print("Calling updateHandler Thread")

            uThread = threading.Thread(target=updateHandler,args=(self.managedClient,devInfo))
            uThread.start()
            uThread.join()
            print("updateHandler Thread is done")


        mgmtRequest = {"action": "firmware/update", "parameters": [{"name": "version", "value": "0.1.11" },
                       {"name": "name", "value": "RasPi01 firmware"}, {"name": "verifier", "value": "123df"},
                       {"name": "uri","value": "https://github.com/ibm-messaging/iot-raspberrypi/releases/download/1.0.2.1/iot_1.0-2_armhf.deb"}],
                       "devices": [{"typeId": self.registeredDevice["typeId"],"deviceId": self.registeredDevice["deviceId"]}]};

        self.managedClient.__firmwareUpdate = device.DeviceFirmware('0.0','0.0','uri','verifier',device.ManagedClient.UPDATESTATE_IDLE,
                                                      device.ManagedClient.UPDATESTATE_IDLE,'updatedDateTime')

        #Setup user defined firmware download call back
        self.managedClient.firmwereActionCallback = firmwareUpdateActionCB
        self.managedClient.connect()
        self.managedClient.setUpdateStatus(device.ManagedClient.UPDATESTATE_IDLE)

        #Initialize device management request for firmware download
        initResult = self.apiClient.initiateDeviceManagementRequest(mgmtRequest)
        reqId = initResult['reqId']
        time.sleep(1)

        #Delete initiated device management request
        assert_true(self.apiClient.deleteDeviceManagementRequest(reqId))

        self.managedClient.disconnect()

    @SkipTest
    def testDMEAction(self):
        def doDMEAction(topic,data,reqId):
            print("In DME Action Callabck")
            print("Received topic = "+topic)
            print("Received reqId = "+reqId)
            print("Received data = %s" %data)
            return True

        self.managedClient.connect()

        dmeData1 = {"bundleId": "example-dme-actions-v1",
                   "displayName": {"en_US": "example-dme Actions v1"},
                   "version": "1.0","actions": {"installPlugin": {
                   "actionDisplayName": { "en_US": "Install Plug-in"},
                   "parameters": [ { "name": "pluginURI",
                                     "value": "http://example.dme.com",
                                    "required": "true" } ] } } }
        dmeData2 = {"bundleId": "example-dme-actions-v2",
                   "displayName": {"en_US": "example-dme Actions v2"},
                   "version": "1.0","actions": {"installPlugin": {
                   "actionDisplayName": { "en_US": "Install Plug-in"},
                   "parameters": [ { "name": "pluginURI",
                                     "value": "http://example.dme.com",
                                    "required": "true" } ] } } }
        dmeData3 = {"bundleId": "example-dme-actions-v3",
                   "displayName": {"en_US": "example-dme Actions v3"},
                   "version": "1.0","actions": {"installPlugin": {
                   "actionDisplayName": { "en_US": "Install Plug-in"},
                   "parameters": [ { "name": "pluginURI",
                                     "value": "http://example.dme.com",
                                    "required": "true" } ] } } }

        self.apiClient.deleteDeviceManagementExtensionPkg('example-dme-actions-v1')
        self.apiClient.deleteDeviceManagementExtensionPkg('example-dme-actions-v2')
        self.apiClient.deleteDeviceManagementExtensionPkg('example-dme-actions-v3')

        addResult = self.apiClient.createDeviceManagementExtensionPkg(dmeData1)
        assert_equal(addResult['bundleId'],'example-dme-actions-v1')
        addResult = self.apiClient.createDeviceManagementExtensionPkg(dmeData2)
        assert_equal(addResult['bundleId'],'example-dme-actions-v2')
        addResult = self.apiClient.createDeviceManagementExtensionPkg(dmeData3)
        assert_equal(addResult['bundleId'],'example-dme-actions-v3')

        self.managedClient.dmeActionCallback = doDMEAction;
        self.managedClient.manage(lifetime=0,supportDeviceMgmtExtActions=True,
                                        bundleIds=['example-dme-actions-v1',
                                                   'example-dme-actions-v2',
                                                   'example-dme-actions-v3'])
        mgmtRequest = {"action": "example-dme-actions-v1/installPlugin",
                       "parameters": [{ "name": "pluginURI",
                                         "value": "http://example.dme.com",}],
                       "devices": [{ "typeId": self.registeredDevice["typeId"], "deviceId": self.registeredDevice["deviceId"] }]}
        initResult = self.apiClient.initiateDeviceManagementRequest(mgmtRequest)
        reqId = initResult['reqId']

        assert_true(self.apiClient.deleteDeviceManagementExtensionPkg('example-dme-actions-v1'))
        assert_true(self.apiClient.deleteDeviceManagementExtensionPkg('example-dme-actions-v2'))
        assert_true(self.apiClient.deleteDeviceManagementExtensionPkg('example-dme-actions-v3'))
        assert_true(self.apiClient.deleteDeviceManagementRequest(reqId))

        self.managedClient.unmanage()
        self.managedClient.disconnect()



# *****************************************************************************
# Copyright (c) 2016 IBM Corporation and other Contributors.
#
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#
# Contributors:
#   Lokesh Haralakatta  - Initial Contribution
# *****************************************************************************

import ibmiotf.gateway
import ibmiotf.application
import uuid
from ibmiotf import *
from nose.tools import *
from nose import SkipTest
import testUtils

class TestGateway(testUtils.AbstractTest):
    registeredDevice = None
    registeredGateway = None
    
    DEVICE_TYPE = "test_device"
    DEVICE_ID = str(uuid.uuid4())
    
    GATEWAY_TYPE = "test_gateway"
    GATEWAY_ID = str(uuid.uuid4())

    
    @classmethod
    def setup_class(self):
        # Register a Device
        try: 
            deviceType = self.setupAppClient.api.getDeviceType(self.DEVICE_TYPE)
        except APIException as e:
            if e.httpCode == 404:
                deviceType = self.setupAppClient.api.addDeviceType(self.DEVICE_TYPE)
            else: 
                raise e
        self.registeredDevice = self.setupAppClient.api.registerDevice(self.DEVICE_TYPE, self.DEVICE_ID)
        
        # Register a Gateway
        try: 
            gatewayType = self.setupAppClient.api.getDeviceType(self.GATEWAY_TYPE)
        except APIException as e:
            if e.httpCode == 404:
                deviceType = self.setupAppClient.api.addDeviceType(self.GATEWAY_TYPE, classId = "Gateway")
            else: 
                raise e
        self.registeredGateway = self.setupAppClient.api.registerDevice(self.GATEWAY_TYPE, self.GATEWAY_ID)
        
        self.options={
            "org": self.ORG_ID,
            "type": self.registeredGateway["typeId"],
            "id": self.registeredGateway["deviceId"],
            "auth-method": "token",
            "auth-token": self.registeredGateway["authToken"]
        }
        

    @classmethod
    def teardown_class(self):
        self.setupAppClient.api.deleteDevice(self.DEVICE_TYPE, self.DEVICE_ID)
        self.setupAppClient.api.deleteDevice(self.GATEWAY_TYPE, self.GATEWAY_ID)


    @raises(Exception)
    def testMissingOptions(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.gateway.Client({})
        assert_equal(e.exception.msg, 'Missing required property: org')

    @raises(Exception)
    def testMissingOrg(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.gateway.Client({"org": None, "type": self.registeredGateway["typeId"], "id": self.registeredGateway["deviceId"], "auth-method": "token", "auth-token": self.registeredGateway["authToken"] })
        assert_equal(e.exception.msg, 'Missing required property: org')

    @raises(Exception)
    def testMissingType(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.gateway.Client({"org": self.ORG_ID, "type": None, "id": self.registeredGateway["deviceId"], "auth-method": "token", "auth-token": self.registeredGateway["authToken"] })
        assert_equal(e.exception.msg, 'Missing required property: type')

    @raises(Exception)
    def testMissingId(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.gateway.Client({"org": self.ORG_ID, "type": self.registeredGateway["typeId"], "id": None, "auth-method": "token", "auth-token": self.registeredGateway["authToken"]})
        assert_equal(e.exception.msg, 'Missing required property: id')

    @raises(Exception)
    def testMissingAuthMethod(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.gateway.Client({"org": self.ORG_ID, "type": self.registeredGateway["typeId"], "id": self.registeredGateway["deviceId"], "auth-method": None, "auth-token": self.registeredGateway["authToken"]})
        assert_equal(e.exception.msg, 'Missing required property: auth-method')

    @raises(Exception)
    def testMissingAuthToken(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.gateway.Client({"org": self.ORG_ID, "type": self.registeredGateway["typeId"], "id": self.registeredGateway["deviceId"], "auth-method": "token", "auth-token": None })
        assert_equal(e.exception.msg, 'Missing required property: auth-token')

    @raises(Exception)
    def testUnSupportedAuthMethod(self):
        with assert_raises(UnsupportedAuthenticationMethod) as e:
            ibmiotf.gateway.Client({"org": self.ORG_ID, "type": self.registeredGateway["typeId"], "id": self.registeredGateway["deviceId"], "auth-method": "unsupported-method", "auth-token": self.registeredGateway["authToken"]})
        assert_equal(e.exception_type,UnsupportedAuthenticationMethod)


    def testGatewayClientInstance(self):
        gatewayCli = ibmiotf.gateway.Client({"org": self.ORG_ID, "type": self.registeredGateway["typeId"], "id": self.registeredGateway["deviceId"], "auth-method": "token", "auth-token": self.registeredGateway["authToken"]})
        assert_is_instance(gatewayCli , ibmiotf.gateway.Client)

    @raises(Exception)
    def testMissingConfigFile(self):
        deviceFile="InvalidFile.out"
        with assert_raises(ConfigurationException) as e:
            ibmiotf.gateway.ParseConfigFile(deviceFile)
        assert_equal(e.exception.msg, 'Error reading device configuration file')

    @raises(Exception)
    def testInvalidConfigFile(self):
        deviceFile="nullValues.conf"
        with assert_raises(AttributeError) as e:
            ibmiotf.gateway.ParseConfigFile(deviceFile)
        assert_equal(e.exception, AttributeError)

    def testNotAuthorizedConnect(self):
        client = ibmiotf.gateway.Client({"org": self.ORG_ID, "type": self.registeredGateway["typeId"], "id": self.registeredGateway["deviceId"], "auth-method": "token", "auth-token": "MGxxx3g7Yjt-6keG(l", "auth-key":"a-xxxxxx-s1tsofmoxo"})
        with assert_raises(ConnectionException) as e:
            client.connect()

    def testMissingMessageEncoder(self):
        gatewayClient = ibmiotf.gateway.Client(self.options)
        gatewayClient.connect()

        with assert_raises(MissingMessageEncoderException) as e:
            myData={'name' : 'foo', 'cpu' : 60, 'mem' : 50}
            gatewayClient.publishDeviceEvent(self.registeredGateway["typeId"],self.registeredGateway["deviceId"],"missingMsgEncode", "jason", myData)

    def testMissingMessageEncoderWithPublishGatewayEvent(self):
        gatewayClient = ibmiotf.gateway.Client(self.options)
        gatewayClient.connect()

        with assert_raises(MissingMessageEncoderException) as e:
            myData={'name' : 'foo', 'cpu' : 60, 'mem' : 50}
            gatewayClient.publishGatewayEvent("missingMsgEncode", "jason", myData)

    def testGatewayPubSubMethods(self):
        gatewayClient = ibmiotf.gateway.Client(self.options)
        gatewayClient.connect()

        def publishCallback():
            print("Publish Event done!!!")

        myData={'name' : 'foo', 'cpu' : 60, 'mem' : 50}
        assert_true(gatewayClient.publishDeviceEvent(self.DEVICE_TYPE, self.DEVICE_ID, "testDevicePublishEventJson", "json", myData, on_publish=publishCallback))
        assert_true(gatewayClient.publishDeviceEvent(self.DEVICE_TYPE, self.DEVICE_ID, "testDevicePublishEventXML", "xml", myData, on_publish=publishCallback))
        assert_true(gatewayClient.publishGatewayEvent("testGatewayPublishEventJson", "json", myData, on_publish=publishCallback))
        assert_true(gatewayClient.publishGatewayEvent("testGatewayPublishEventXML", "xml", myData, on_publish=publishCallback))

        assert_true(gatewayClient.subscribeToDeviceCommands(self.DEVICE_TYPE, self.DEVICE_ID))
        assert_true(gatewayClient.subscribeToGatewayCommands())
        assert_true(gatewayClient.subscribeToGatewayNotifications())

        gatewayClient.disconnect()

    def testGatewayPubSubMethodsInQSMode(self):
        gatewayClient = ibmiotf.gateway.Client({"org": "quickstart", "type": self.registeredGateway["typeId"], "id": self.registeredGateway["deviceId"], "auth-method":"None", "auth-token":"None" })
        gatewayClient.connect()

        def publishCallback():
            print("Publish Event done!!!")

        myData={'name' : 'foo', 'cpu' : 60, 'mem' : 50}
        
        assert_true(gatewayClient.publishDeviceEvent(self.DEVICE_TYPE, self.DEVICE_ID, "testDevicePublishEvent", "json", myData, on_publish=publishCallback))
        assert_true(gatewayClient.publishGatewayEvent("testGatewayPublishEvent", "json", myData, on_publish=publishCallback))

        assert_false(gatewayClient.subscribeToDeviceCommands(self.DEVICE_TYPE, self.DEVICE_ID))
        assert_false(gatewayClient.subscribeToGatewayCommands())
        assert_false(gatewayClient.subscribeToGatewayNotifications())

        gatewayClient.disconnect()

    def testDeviceInfoInstance(self):
        deviceInfoObj = ibmiotf.gateway.DeviceInfo()
        assert_is_instance(deviceInfoObj, ibmiotf.gateway.DeviceInfo)
        print(deviceInfoObj)
    
    @SkipTest
    def testPublishCommandByApplication(self):
        def deviceCmdCallback(cmd):
            assert_true(cmd.data['rebootDelay'] == 50)

        def gatewayCmdCallback(cmd):
            assert_true(cmd.data['rebootDelay'] == 50)

        def notificationCallback(cmd):
            assert_true(cmd.data['rebootDelay'] == 50)

        def appCmdPublishCallback():
            print("Application Publish Command done!!!")

        gatewayClient = ibmiotf.gateway.Client(self.options)
        
        gatewayClient.commandCallback = gatewayCmdCallback
        gatewayClient.deviceCommandCallback = deviceCmdCallback
        gatewayClient.notificationCallback = notificationCallback
        gatewayClient.connect()
        gatewayClient.subscribeToDeviceCommands(self.DEVICE_TYPE, self.DEVICE_ID)
        gatewayClient.subscribeToGatewayCommands()
        gatewayClient.subscribeToGatewayNotifications()

        appClient = ibmiotf.application.Client(self.appOptions)
        appClient.connect()

        commandData={'rebootDelay' : 50}
        
        assert_true(appClient.publishCommand(self.registeredGateway["typeId"], self.registeredGateway["deviceId"], "reboot", "json", commandData, on_publish=appCmdPublishCallback))
        time.sleep(2)
        
        assert_true(appClient.publishCommand(self.DEVICE_TYPE, self.DEVICE_ID, "reboot", "json", commandData, on_publish=appCmdPublishCallback))
        time.sleep(2)

        appClient.disconnect()
        gatewayClient.disconnect()

    def testGatewayApiClientSupport(self):
        gatewayClient = ibmiotf.gateway.Client(self.options)
        assert_is_instance(gatewayClient.api, ibmiotf.api.ApiClient)

        #Add new device
        newDeviceId = str(uuid.uuid4())
        addResult = gatewayClient.api.registerDevice(self.DEVICE_TYPE, newDeviceId)
        
        assert_equal(addResult['typeId'], self.DEVICE_TYPE)
        assert_equal(addResult['deviceId'], newDeviceId)

        #Remove the added device
        gatewayClient.api.deleteDevice(self.DEVICE_TYPE, newDeviceId)



'''
This is the default encoder used by clients for all messages sent with format 
defined as "json-iotf".  This default can be changed by reconfiguring your client:
  
  deviceCli.setMessageEncoderModule("json-iotf", myCustomEncoderModule)

'''
import json
from datetime import datetime
import pytz
import iso8601
from ibmiotf import Message, InvalidEventException


'''
Convert a Python dictionary object into a UTF-8 encoded JSON string. The data is placed 
inside a top levle "d" element with a timestamp added to "ts".

* The entire message is converted to JSON
* The top level d element is considered as the message data
* The timestamp of the message is set to the value read from the ts element
* If no ts element is present the timestamp defaults to the time the message is received
'''
def encode(data=None, timestamp=None):
    ts = timestamp.isoformat()
    
    payload = { 'd': data, 'ts': ts }
    return json.dumps(payload)

'''
Convert a JSON message where all data is under a top level "d" element and an optional "ts" element exists, containing a timestamp for the message.

* The entire message is converted to JSON
* The top level d element is considered as the message data
* The timestamp of the message is set to the value read from the ts element
* If no ts element is present the timestamp defaults to the time the message is received
'''
def decode(message):
    try:
        payload = json.loads(message.payload.decode("utf-8"))
    except ValueError as e:
        raise InvalidEventException("Unable to parse JSON.  payload=\"%s\" error=%s" % (message.payload, str(e)))
        
    data = None
    timestamp = None
    
    # Try to parse a timestamp
    try:
        if 'ts' in payload:
            dt = iso8601.parse_date(payload['ts'])
            timestamp = dt.astimezone(pytz.timezone('UTC'))
        else:
            timestamp = datetime.now(pytz.timezone('UTC'))
    except iso8601.ParseError as e:
        raise InvalidEventException("Unable to parse event timestamp: %s" % str(e))

    # Try to parse the data
    if 'd' in payload:
        data = payload['d']
        
        # TODO: Flatten JSON, covert into array of key/value pairs
    
    return Message(data, timestamp)



# *****************************************************************************
# Copyright (c) 2016 IBM Corporation and other Contributors.
#
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#
# Contributors:
#   Lokesh Haralakatta  - Initial Contribution
# *****************************************************************************

import ibmiotf.device
import ibmiotf.application
import uuid
import os
from ibmiotf import *
from nose.tools import *
from nose import SkipTest
import logging
import testUtils

class TestDevice(testUtils.AbstractTest):
    registeredDevice = None
    httpClient = None

    DEVICE_TYPE = "test_device"
    DEVICE_ID = str(uuid.uuid4())
    
    @classmethod
    def setup_class(self):
        try: 
            deviceType = self.setupAppClient.api.getDeviceType(self.DEVICE_TYPE)
        except APIException as e:
            if e.httpCode == 404:
                deviceType = self.setupAppClient.api.addDeviceType(self.DEVICE_TYPE)
            else: 
                raise e
        
        self.registeredDevice = self.setupAppClient.api.registerDevice(self.DEVICE_TYPE, self.DEVICE_ID)
        
        self.options={
            "org": self.ORG_ID,
            "type": self.registeredDevice["typeId"],
            "id": self.registeredDevice["deviceId"],
            "auth-method": "token",
            "auth-token": self.registeredDevice["authToken"]
        }
        
        self.httpClient = ibmiotf.device.HttpClient(self.options)
    

    @classmethod
    def teardown_class(self):
        del self.httpClient
        self.setupAppClient.api.deleteDevice(self.DEVICE_TYPE, self.DEVICE_ID)
    
    
    def testPublishEventOverHTTPs(self):
        myData={'name' : 'foo', 'cpu' : 60, 'mem' : 50}
        assert_equals(self.httpClient.publishEvent("testPublishEventHTTPs", "json",myData),200)
        assert_equals(self.httpClient.publishEvent("testPublishEventHTTPs", "xml",myData),200)
    
    
    def testPublishEventOverHTTP(self):
        client = ibmiotf.device.HttpClient({"org": "quickstart", "type": self.registeredDevice["typeId"], "id": self.registeredDevice["deviceId"],
                                        "auth-method":"None", "auth-token":"None" })
        myData={'name' : 'foo', 'cpu' : 60, 'mem' : 50}
        assert_equals(client.publishEvent("testPublishEventHTTP", "json",myData),200)
        assert_equals(client.publishEvent("testPublishEventHTTP", "xml",myData),200)
    


# *****************************************************************************
# Copyright (c) 2015, 2017 IBM Corporation and other Contributors.
#
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#
# Contributors:
#   David Parker
#   Paul Slater
#   Amit M Mangalvedkar
#   Lokesh K Haralakatta
#   Ian Craggs
# *****************************************************************************

import ibmiotf
import json
import requests
import iso8601
import base64
import json
from datetime import datetime

import logging
from symbol import parameters
from requests_toolbelt.multipart.encoder import MultipartEncoder


class ApiClient():
    #Organization URL
    organizationUrl = 'https://%s/api/v0002/'

    #Bulk Operations URL
    bulkRetrieve = 'https://%s/api/v0002/bulk/devices'
    bulkAddUrl = 'https://%s/api/v0002/bulk/devices/add'
    bulkRemoveUrl = 'https://%s/api/v0002/bulk/devices/remove'

    #Device Types URL
    deviceTypesUrl = 'https://%s/api/v0002/device/types'
    deviceTypeUrl = 'https://%s/api/v0002/device/types/%s'

    #Device URL
    devicesUrl = 'https://%s/api/v0002/device/types/%s/devices'
    deviceUrl = 'https://%s/api/v0002/device/types/%s/devices/%s'
    deviceUrlLocation = 'https://%s/api/v0002/device/types/%s/devices/%s/location'
    deviceUrlMgmt = 'https://%s/api/v0002/device/types/%s/devices/%s/mgmt'

    #Device Event Cache URLs
    deviceEventListCacheUrl = 'https://%s/api/v0002/device/types/%s/devices/%s/events'
    deviceEventCacheUrl = 'https://%s/api/v0002/device/types/%s/devices/%s/events/%s'

    #Log Events URL
    deviceLogs = 'https://%s/api/v0002/logs/connection'

    #Diagnostics URL
    deviceDiagLogs = 'https://%s/api/v0002/device/types/%s/devices/%s/diag/logs'
    deviceDiagLogsLogId = 'https://%s/api/v0002/device/types/%s/devices/%s/diag/logs/%s'
    deviceDiagErrorCodes = 'https://%s/api/v0002/device/types/%s/devices/%s/diag/errorCodes'

    #Usage Management URL
    usageMgmt = 'https://%s/api/v0002/usage'

    #Service Status URL
    serviceStatus = 'https://%s/api/v0002/service-status'

    #Device Management URL
    mgmtRequests = 'https://%s/api/v0002/mgmt/requests'
    mgmtSingleRequest = 'https://%s/api/v0002/mgmt/requests/%s'
    mgmtRequestStatus = 'https://%s/api/v0002/mgmt/requests/%s/deviceStatus'
    mgmtRequestSingleDeviceStatus = 'https://%s/api/v0002/mgmt/requests/%s/deviceStatus/%s/%s'

    #Device Management Extensions (dme) URL
    dmeRequests = 'https://%s/api/v0002/mgmt/custom/bundle'
    dmeSingleRequest = 'https://%s/api/v0002/mgmt/custom/bundle/%s'

    # Information management URLs

    # Draft Device type URLs
    draftDeviceTypeUrl = 'https://%s/api/v0002/draft/device/types/%s'

    # Schema URLs
    allSchemasUrl = "https://%s/api/v0002%s/schemas"
    oneSchemaUrl  = "https://%s/api/v0002%s/schemas/%s"
    oneSchemaContentUrl  = "https://%s/api/v0002%s/schemas/%s/content"

    # Event type URLs
    allEventTypesUrl = "https://%s/api/v0002%s/event/types"
    oneEventTypeUrl  = "https://%s/api/v0002%s/event/types/%s"

    # Physical Interface URLs
    allPhysicalInterfacesUrl = "https://%s/api/v0002%s/physicalinterfaces"
    onePhysicalInterfaceUrl  = "https://%s/api/v0002%s/physicalinterfaces/%s"
    oneDeviceTypePhysicalInterfaceUrl = 'https://%s/api/v0002/draft/device/types/%s/physicalinterface'

    # Event URLs
    allEventsUrl = "https://%s/api/v0002%s/physicalinterfaces/%s/events"
    oneEventUrl  = "https://%s/api/v0002%s/physicalinterfaces/%s/events/%s"

    # Logical Interface URLs
    allLogicalInterfacesUrl = "https://%s/api/v0002%s/logicalinterfaces"
    oneLogicalInterfaceUrl  = "https://%s/api/v0002%s/logicalinterfaces/%s"
    allDeviceTypeLogicalInterfacesUrl = "https://%s/api/v0002%s/device/types/%s/logicalinterfaces"
    oneDeviceTypeLogicalInterfaceUrl = "https://%s/api/v0002/draft/device/types/%s/logicalinterfaces/%s"

    # Mappings
    allDeviceTypeMappingsUrl = "https://%s/api/v0002%s/device/types/%s/mappings"
    oneDeviceTypeMappingUrl = "https://%s/api/v0002%s/device/types/%s/mappings/%s"

    # Device state
    deviceStateUrl = "https://%s/api/v0002/device/types/%s/devices/%s/state/%s"

    def __init__(self, options, logger=None):
        self.__options = options

        # Configure logging
        if logger is None:
            logger = logging.getLogger(self.__module__+"."+self.__class__.__name__)
            logger.setLevel(logging.INFO)

        self.logger = logger

        if 'auth-key' not in self.__options or self.__options['auth-key'] is None:
            raise ibmiotf.ConfigurationException("Missing required property for API key based authentication: auth-key")
        if 'auth-token' not in self.__options or self.__options['auth-token'] is None:
            raise ibmiotf.ConfigurationException("Missing required property for API key based authentication: auth-token")

        # Get the orgId from the apikey
        self.__options['org'] = self.__options['auth-key'][2:8]

        if "domain" not in self.__options:
            # Default to the domain for the public cloud offering
            self.__options['domain'] = "internetofthings.ibmcloud.com"

        self.host = self.__options['org'] + "." + self.__options['domain']
        self.credentials = (self.__options['auth-key'], self.__options['auth-token'])

        # To support development systems this can be overridden to False
        self.verify = False
        if not self.verify:
            from requests.packages.urllib3.exceptions import InsecureRequestWarning
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

    def deleteDevice(self, typeId, deviceId):
        """
        Delete an existing device.
        It accepts typeId (string) and deviceId (string) as parameters
        In case of failure it throws APIException
        """
        deviceUrl = ApiClient.deviceUrl % (self.host, typeId, deviceId)

        r = requests.delete(deviceUrl, auth=self.credentials, verify=self.verify)
        status = r.status_code
        if status == 204:
            self.logger.debug("Device was successfully removed")
            return True
        elif status == 401:
            raise ibmiotf.APIException(401, "The authentication token is empty or invalid", None)
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(status, "Unexpected error", None)




    #This method returns the organization
    def getOrganizationDetails(self):
        """
        Get details about an organization
        It does not need any parameter to be passed
        In case of failure it throws APIException
        """
        if self.__options['org'] is None:
            raise ibmiotf.ConfigurationException("Missing required property: org")
        else:
            url = ApiClient.organizationUrl % (self.host)
        r = requests.get(url, auth=self.credentials, verify=self.verify)
        status = r.status_code
        if status == 200:
            self.logger.debug("Organization retrieved")
            return r.json()
        elif status == 401:
            raise ibmiotf.APIException(401, "The authentication token is empty or invalid", None)
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "The organization does not exist", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def getDevices(self, parameters = None):
        """
        Retrieve bulk devices
        It accepts accepts a list of parameters
        In case of failure it throws APIException
        """
        bulkRetrieve = ApiClient.bulkRetrieve % (self.host )
        r = requests.get(bulkRetrieve, auth = self.credentials, params = parameters, verify=self.verify)

        status = r.status_code

        if status == 200:
            self.logger.debug("Bulk retrieval successful")
            return r.json()
        elif status == 401:
            raise ibmiotf.APIException(401, "The authentication token is empty or invalid", None)
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the API key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "The organization or device type does not exist", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def registerDevices(self, listOfDevices):
        """
        Register multiple new devices, each request can contain a maximum of 512KB.
        The response body will contain the generated authentication tokens for all devices.
        You must make sure to record these tokens when processing the response.
        We are not able to retrieve lost authentication tokens
        It accepts accepts a list of devices (List of Dictionary of Devices)
        In case of failure it throws APIException
        """
        bulkAdd = ApiClient.bulkAddUrl % (self.host )
        r = requests.post(bulkAdd, auth = self.credentials, data = json.dumps(listOfDevices), headers = {'content-type': 'application/json'}, verify=self.verify)

        status = r.status_code

        if status == 201:
            self.logger.debug("Bulk registration successful")
            return r.json()
        elif status == 202:
            raise ibmiotf.APIException(202, "Partial Success. Some devices registered successfully", r.json())
        elif status == 400:
            raise ibmiotf.APIException(400, "Invalid request (No body, invalid JSON, unexpected key, bad value)", r.json())
        elif status == 403:
            raise ibmiotf.APIException(403, "Maximum number of devices exceeded", r.json())
        elif status == 413:
            raise ibmiotf.APIException(413, "Request content exceeds 512KB", r.json())
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def deleteMultipleDevices(self, listOfDevices):
        """
        Delete multiple devices, each request can contain a maximum of 512Kb
        It accepts accepts a list of devices (List of Dictionary of Devices)
        In case of failure it throws APIException
        """
        bulkRemove = ApiClient.bulkRemoveUrl % (self.host )
        r = requests.post(bulkRemove, auth = self.credentials, data = json.dumps(listOfDevices), headers = {'content-type': 'application/json'}, verify=self.verify)

        status = r.status_code
        if status == 202:
            self.logger.debug("Some devices deleted successfully")
            return r.json()
        elif status == 400:
            raise ibmiotf.APIException(400, "Invalid request (No body, invalid JSON, unexpected key, bad value)", r.json())
        elif status == 413:
            raise ibmiotf.APIException(413, "Request content exceeds 512KB", r.json())
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)



    def getDeviceTypes(self, parameters = None):
        """
        Retrieves all existing device types.
        It accepts accepts an optional query parameters (Dictionary)
        In case of failure it throws APIException
        """
        deviceTypeUrl = ApiClient.deviceTypesUrl % (self.host)
        r = requests.get(deviceTypeUrl, auth=self.credentials, params = parameters, verify=self.verify)
        status = r.status_code
        if status == 200:
            self.logger.debug("Device types successfully retrieved")
            self.logger.debug(json.dumps(r.json()))
            return r.json()
        elif status == 401:
            raise ibmiotf.APIException(401, "The authentication token is empty or invalid", None)
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)



    def addDeviceType(self, typeId, description = None, deviceInfo = None, metadata = None, classId = "Device"):
        """
        Creates a device type.
        It accepts typeId (string), description (string), deviceInfo(dict) and metadata(dict) as parameter
        In case of failure it throws APIException
        """
        deviceTypesUrl = ApiClient.deviceTypesUrl % (self.host)
        payload = {'id' : typeId, 'description' : description, 'deviceInfo' : deviceInfo, 'metadata': metadata,'classId': classId}

        r = requests.post(deviceTypesUrl, auth=self.credentials, data=json.dumps(payload), headers = {'content-type': 'application/json'}, verify=self.verify)
        status = r.status_code
        if status == 201:
            self.logger.debug("Device Type Created")
            return r.json()
        elif status == 400:
            raise ibmiotf.APIException(400, "Invalid request (No body, invalid JSON, unexpected key, bad value)", r.json())
        elif status == 401:
            raise ibmiotf.APIException(401, "The authentication token is empty or invalid", None)
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 409:
            raise ibmiotf.APIException(403, "The device type already exists", r.json())
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def deleteDeviceType(self, typeId):
        """
        Deletes a device type.
        It accepts typeId (string) as the parameter
        In case of failure it throws APIException
        """
        deviceTypeUrl = ApiClient.deviceTypeUrl % (self.host, typeId)

        r = requests.delete(deviceTypeUrl, auth=self.credentials, verify=self.verify)
        status = r.status_code
        if status == 204:
            self.logger.debug("Device type was successfully deleted")
            return True
        elif status == 401:
            raise ibmiotf.APIException(401, "The authentication token is empty or invalid", None)
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)

    def getDeviceType(self, typeId):
        """
        Gets device type details.
        It accepts typeId (string) as the parameter
        In case of failure it throws APIException
        """
        deviceTypeUrl = ApiClient.deviceTypeUrl % (self.host, typeId)
        r = requests.get(deviceTypeUrl, auth=self.credentials, verify=self.verify)
        status = r.status_code
        if status == 200:
            self.logger.debug("Device type was successfully retrieved")
            return r.json()
        elif status == 401:
            raise ibmiotf.APIException(401, "The authentication token is empty or invalid", None)
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "The device type does not exist", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def updateDeviceType(self, typeId, description, deviceInfo, metadata = None):
        """
        Updates a device type.
        It accepts typeId (string), description (string), deviceInfo (JSON) and metadata(JSON) as the parameters
        In case of failure it throws APIException
        """
        deviceTypeUrl = ApiClient.deviceTypeUrl % (self.host, typeId)
        deviceTypeUpdate = {'description' : description, 'deviceInfo' : deviceInfo, 'metadata' : metadata}
        r = requests.put(deviceTypeUrl, auth=self.credentials, data=json.dumps(deviceTypeUpdate), headers = {'content-type': 'application/json'}, verify=self.verify)
        status = r.status_code
        if status == 200:
            self.logger.debug("Device type was successfully modified")
            return r.json()
        elif status == 401:
            raise ibmiotf.APIException(401, "The authentication token is empty or invalid", None)
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "The device type does not exist", None)
        elif status == 409:
            raise ibmiotf.APIException(409, "The update could not be completed due to a conflict", r.json())
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)

    """
    ===========================================================================
    Device API methods
     - register a new device
     - get a single device
     - remove device
     - update device
    ===========================================================================
    """

    def registerDevice(self, typeId, deviceId, authToken = None, deviceInfo = None, location = None, metadata=None):
        """
        Registers a new device.
        It accepts typeId (string), deviceId (string), authToken (string), location (JSON) and metadata (JSON) as parameters
        In case of failure it throws APIException
        """
        devicesUrl = ApiClient.devicesUrl % (self.host, typeId)
        payload = {'deviceId' : deviceId, 'authToken' : authToken, 'deviceInfo' : deviceInfo, 'location' : location, 'metadata': metadata}

        r = requests.post(devicesUrl, auth=self.credentials, data=json.dumps(payload), headers = {'content-type': 'application/json'}, verify=self.verify)
        status = r.status_code
        if status == 201:
            self.logger.debug("Device Instance Created")
            return r.json()
        elif status == 400:
            raise ibmiotf.APIException(400, "Invalid request (No body, invalid JSON, unexpected key, bad value)", r.json())
        elif status == 401:
            raise ibmiotf.APIException(401, "The authentication token is empty or invalid", None)
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 409:
            raise ibmiotf.APIException(403, "The device already exists", r.json())
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def getDevice(self, typeId, deviceId, expand = None):
        """
        Gets device details.
        It accepts typeId (string), deviceId (string) and expand (JSON) as parameters
        In case of failure it throws APIException
        """
        deviceUrl = ApiClient.deviceUrl % (self.host, typeId, deviceId)

        r = requests.get(deviceUrl, auth=self.credentials, params = expand, verify=self.verify)
        status = r.status_code

        if status == 200:
            self.logger.debug("Device was successfully retrieved")
            return r.json()
        elif status == 401:
            raise ibmiotf.APIException(401, "The authentication token is empty or invalid", None)
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "The device does not exist", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def getDevicesForType(self, typeId, parameters = None):
        """
        Gets details for multiple devices of a type
        It accepts typeId (string), deviceId (string) and expand (JSON) as parameters
        In case of failure it throws APIException
        """
        deviceUrl = ApiClient.devicesUrl % (self.host, typeId)

        r = requests.get(deviceUrl, auth=self.credentials, params = parameters, verify=self.verify)
        status = r.status_code
        if status == 200:
            self.logger.debug("Device was successfully retrieved")
            return r.json()
        elif status == 401:
            raise ibmiotf.APIException(401, "The authentication token is empty or invalid", None)
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "The device does not exist", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)



    def removeDevice(self, typeId, deviceId):
        """
        Delete an existing device.
        It accepts typeId (string) and deviceId (string) as parameters
        In case of failure it throws APIException
        """
        deviceUrl = ApiClient.deviceUrl % (self.host, typeId, deviceId)

        r = requests.delete(deviceUrl, auth=self.credentials, verify=self.verify)
        status = r.status_code
        if status == 204:
            self.logger.debug("Device was successfully removed")
            return True
        elif status == 401:
            raise ibmiotf.APIException(401, "The authentication token is empty or invalid", None)
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def updateDevice(self, typeId, deviceId, metadata, deviceInfo = None, status = None):
        """
        Updates a device.
        It accepts typeId (string), deviceId (string), metadata (JSON), deviceInfo (JSON) and status(JSON) as parameters
        In case of failure it throws APIException
        """
        deviceUrl = ApiClient.deviceUrl % (self.host, typeId, deviceId)

        payload = {'status' : status, 'deviceInfo' : deviceInfo, 'metadata': metadata}
        r = requests.put(deviceUrl, auth=self.credentials, data=json.dumps(payload), headers = {'content-type': 'application/json'}, verify=self.verify)

        status = r.status_code
        if status == 200:
            self.logger.debug("Device was successfully modified")
            return r.json()
        elif status == 401:
            raise ibmiotf.APIException(401, "The authentication token is empty or invalid", None)
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "The organization, device type or device does not exist", None)
        elif status == 409:
            raise ibmiotf.APIException(409, "The update could not be completed due to a conflict", r.json())
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    """
    ===========================================================================
    Last Event Cache Methods
     - get event(s) from cache for device
    ===========================================================================
    """

    def getLastEvent(self, typeId, deviceId, eventId):
        """
        Retrieves Last Cached Event.
        """
        events = ApiClient.deviceEventCacheUrl % (self.host, typeId, deviceId, eventId)
        r = requests.get(events, auth=self.credentials, verify=self.verify)

        status = r.status_code
        if status == 200:
            response = r.json()
            if response["format"] == "json":
                # Convert from base64 to byte to string to dictionary
                jsonPayload = json.loads(base64.b64decode(response["payload"]))
                response["data"] = jsonPayload
            return response

        elif status == 404:
            raise ibmiotf.APIException(404, "Event not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def getLastEvents(self, typeId, deviceId):
        """
        Retrieves all last cached events
        """
        events = ApiClient.deviceEventListCacheUrl % (self.host, typeId, deviceId)
        r = requests.get(events, auth=self.credentials, verify=self.verify)

        status = r.status_code
        if status == 200:
            response = r.json()
            for event in response:
                if event["format"] == "json":
                    # Convert from base64 to byte to string to dictionary
                    jsonPayload = json.loads(base64.b64decode(event["payload"]))
                    event["data"] = jsonPayload
            return response

        elif status == 404:
            raise ibmiotf.APIException(404, "Event not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    """
    ===========================================================================
    Extended Device Model Methods
     - get location for device
     - update location for device
     - get device management information
    ===========================================================================
    """

    def getDeviceLocation(self, typeId, deviceId):
        """
        Retrieve Device Location.
        It accepts typeId (string) and deviceId (string) as parameters
        In case of failure it throws APIException
        """
        deviceUrl = ApiClient.deviceUrlLocation % (self.host, typeId, deviceId)

        r = requests.get(deviceUrl, auth=self.credentials, verify=self.verify)
        status = r.status_code
        if status == 200:
            self.logger.debug("Device Location was successfully obtained")
            return r.json()
        elif status == 404:
            raise ibmiotf.APIException(404, "Device location information not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def updateDeviceLocation(self, typeId, deviceId, deviceLocation):
        """
        Updates the location information for a device. If no date is supplied, the entry is added with the current date and time.
        It accepts typeId (string), deviceId (string) and deviceLocation (JSON) as parameters
        In case of failure it throws APIException
        """
        deviceUrl = ApiClient.deviceUrlLocation % (self.host, typeId, deviceId)

        r = requests.put(deviceUrl, auth=self.credentials, data=json.dumps(deviceLocation), headers = {'content-type': 'application/json'}, verify=self.verify)
        status = r.status_code
        if status == 200:
            self.logger.debug("Device Location was successfully modified")
            return r.json()
        elif status == 404:
            raise ibmiotf.APIException(404, "Device location information not found", None)
        elif status == 409:
            raise ibmiotf.APIException(404, "The update could not be completed due to a conflict", r.json())
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def getDeviceManagementInformation(self, typeId, deviceId):
        """
        Gets device management information for a device.
        It accepts typeId (string) and deviceId (string) as parameters
        In case of failure it throws APIException
        """
        deviceUrl = ApiClient.deviceUrlMgmt % (self.host, typeId, deviceId)
        r = requests.get(deviceUrl, auth=self.credentials, verify=self.verify)
        status = r.status_code
        if status == 200:
            self.logger.debug("Device Management Information was successfully obtained")
            return r.json()
        #This also throws a 403, which has not been documented
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "Device not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def getConnectionLogs(self, parameters):
        """
        List connection log events for a device to aid in diagnosing connectivity problems.
        The entries record successful connection, unsuccessful connection attempts, intentional disconnection and server-initiated disconnection.
        It accepts typeId (string) and deviceId (string) as parameters
        In case of failure it throws APIException
        """
        logs = ApiClient.deviceLogs % (self.host)
        r = requests.get(logs, auth=self.credentials, params = parameters, verify=self.verify)
        status = r.status_code
        if status == 200:
            self.logger.debug("Connection Logs were successfully obtained")
            return r.json()
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "Device not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)

    """
    ===========================================================================
    Device Diagnostics - Logs
     - get logs
     - clear logs
     - add log
     - get log
    ===========================================================================
    """

    def getAllDiagnosticLogs(self, typeId, deviceId):
        """
        Retrieves All Device Diagnostic Logs.
        It accepts typeId (string) and deviceId (string) as parameters
        In case of failure it throws APIException
        """
        deviceDiagnostics = ApiClient.deviceDiagLogs % (self.host, typeId, deviceId)
        r = requests.get(deviceDiagnostics, auth=self.credentials, verify=self.verify)
        status = r.status_code
        if status == 200:
            self.logger.debug("All Diagnostic logs successfully retrieved")
            return r.json()
        elif status == 404:
            raise ibmiotf.APIException(404, "Device not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)



    def clearAllDiagnosticLogs(self, typeId, deviceId):
        """
        Deletes All Device Diagnostic Logs.
        It accepts typeId (string) and deviceId (string) as parameters
        In case of failure it throws APIException
        """
        deviceDiagnostics = ApiClient.deviceDiagLogs % (self.host, typeId, deviceId)
        r = requests.delete(deviceDiagnostics, auth=self.credentials, verify=self.verify)
        status = r.status_code
        if status == 204:
            self.logger.debug("All Diagnostic logs successfully cleared")
            return True
        #403 and 404 error code needs to be added in Swagger documentation
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "Device not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def addDiagnosticLog(self, typeId, deviceId, logs):
        """
        Add Device Diagnostic Logs.
        It accepts typeId (string), deviceId (string) and logs (JSON) as parameters
        In case of failure it throws APIException
        """
        deviceDiagnostics = ApiClient.deviceDiagLogs % (self.host, typeId, deviceId)
        r = requests.post(deviceDiagnostics, auth=self.credentials, data = json.dumps(logs), headers = {'content-type': 'application/json'}, verify=self.verify)

        status = r.status_code
        if status == 201:
            self.logger.debug("Diagnostic entry was successfully added")
            return True
        #403 and 404 error code needs to be added in Swagger documentation
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "Device not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def getDiagnosticLog(self, typeId, deviceId, logId):
        """
        Retrieves Device Diagnostic Logs.
        It accepts typeId (string), deviceId (string) and logId (string) as parameters
        In case of failure it throws APIException
        """
        deviceDiagnostics = ApiClient.deviceDiagLogsLogId % (self.host, typeId, deviceId, logId)
        r = requests.get(deviceDiagnostics, auth=self.credentials, verify=self.verify)
        status = r.status_code
        if status == 200:
            self.logger.debug("Diagnostic log successfully retrieved")
            return r.json()
        elif status == 404:
            raise ibmiotf.APIException(404, "Device not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def clearDiagnosticLog(self, typeId, deviceId, logId):
        """
        Delete Device Diagnostic Logs.
        It accepts typeId (string), deviceId (string) and logId (string) as parameters
        In case of failure it throws APIException
        """
        deviceDiagnostics = ApiClient.deviceDiagLogsLogId % (self.host, typeId, deviceId, logId)
        r = requests.delete(deviceDiagnostics, auth=self.credentials, verify=self.verify)
        status = r.status_code
        if status == 204:
            self.logger.debug("Diagnostic log successfully cleared")
            return True
        #403 and 404 error code needs to be added in Swagger documentation
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "Device not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    """
    ===========================================================================
    Device Diagnostics - Error Codes
     - add error code
     - get all error codes
     - clear all error codes
    ===========================================================================
    """

    def addErrorCode(self, typeId, deviceId, errorCode):
        """
        Adds an error code to the list of error codes for the device. The list may be pruned as the new entry is added.
        It accepts typeId (string), deviceId (string) and errorCode (JSON) as parameters
        In case of failure it throws APIException
        """
        deviceDiagnostics = ApiClient.deviceDiagErrorCodes % (self.host, typeId, deviceId)
        r = requests.post(deviceDiagnostics, auth=self.credentials, data = json.dumps(errorCode), headers = {'content-type': 'application/json'}, verify=self.verify)

        status = r.status_code
        if status == 201:
            self.logger.debug("Error code was successfully added")
            return True
        #403 and 404 error code needs to be added in Swagger documentation
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "Error Code not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def getAllDiagnosticErrorCodes(self, typeId, deviceId):
        """
        Gets diagnostic error codes for a device.
        It accepts typeId (string) and deviceId (string) as parameters
        In case of failure it throws APIException
        """
        deviceDiagnostics = ApiClient.deviceDiagErrorCodes % (self.host, typeId, deviceId)
        r = requests.get(deviceDiagnostics, auth=self.credentials, verify=self.verify)

        status = r.status_code
        if status == 200:
            self.logger.debug("Error codes were successfully retrieved")
            return r.json()
        #403 and 404 error code needs to be added in Swagger documentation
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "Error Code not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def clearAllErrorCodes(self, typeId, deviceId):
        """
        Clears the list of error codes for the device. The list is replaced with a single error code of zero.
        It accepts typeId (string) and deviceId (string) as parameters
        In case of failure it throws APIException
        """
        deviceDiagnostics = ApiClient.deviceDiagErrorCodes % (self.host, typeId, deviceId)
        r = requests.delete(deviceDiagnostics, auth=self.credentials, verify=self.verify)

        status = r.status_code
        if status == 204:
            self.logger.debug("Error codes successfully cleared")
            return True
        #403 and 404 error code needs to be added in Swagger documentation
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "Error Code not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    """
    ===========================================================================
    Service Status API
    - Retrieve service status
    ===========================================================================
    """

    def getServiceStatus(self):
        """
        Retrieve the organization-specific status of each of the services offered by the IBM Watson IoT Platform.
        In case of failure it throws APIException
        """
        serviceStatus = ApiClient.serviceStatus % (self.host)
        r = requests.get(serviceStatus, auth=self.credentials, verify=self.verify)

        status = r.status_code

        if status == 200:
            self.logger.debug("Service status successfully retrieved")
            return r.json()
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)

    """
    ===========================================================================
    Usage API
    - Active Devices
    - Data Traffic
    - Storage
    ===========================================================================
    """

    def getActiveDevices(self, options):
        """
        Retrieve the number of active devices over a period of time.
        In case of failure it throws APIException
        """
        activeDevices = (ApiClient.usageMgmt + '/active-devices') % (self.host)
        r = requests.get(activeDevices, auth=self.credentials, params=options, verify=self.verify)

        status = r.status_code

        if status == 200:
            self.logger.debug("Active Devices = ", r.json() )
            return r.json()
        elif status == 400:
            raise ibmiotf.APIException(400, "Bad Request", r.json())
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def getDataTraffic(self, options):
        """
        Retrieve the amount of data used.
        In case of failure it throws APIException
        """
        dataTraffic = (ApiClient.usageMgmt + '/data-traffic') % (self.host)
        r = requests.get(dataTraffic, auth=self.credentials, params=options, verify=self.verify)

        status = r.status_code

        if status == 200:
            self.logger.debug("Data Traffic = ", r.json() )
            return r.json()
        elif status == 400:
            raise ibmiotf.APIException(400, "Bad Request", r.json())
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    """
    ===========================================================================
    Device Management API
    - Get all requests
    - Initiate new request
    - Delete request
    - Get request
    - Get request status
    - Get request status for specific device
    ===========================================================================
    """

    def getAllDeviceManagementRequests(self):
        """
        Gets a list of device management requests, which can be in progress or recently completed.
        In case of failure it throws APIException
        """
        mgmtRequests = ApiClient.mgmtRequests % (self.host)
        r = requests.get(mgmtRequests, auth=self.credentials, verify=self.verify)

        status = r.status_code

        if status == 200:
            self.logger.debug("Retrieved all device management requests")
            return r.json()
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def initiateDeviceManagementRequest(self, deviceManagementRequest):
        """
        Initiates a device management request, such as reboot.
        In case of failure it throws APIException
        """
        mgmtRequests = ApiClient.mgmtRequests % (self.host)
        r = requests.post(mgmtRequests, auth=self.credentials, data=json.dumps(deviceManagementRequest), headers = {'content-type': 'application/json'}, verify=self.verify)

        status = r.status_code
        if status == 202:
            self.logger.debug("The request has been accepted for processing")
            return r.json()
        elif status == 403:
            raise ibmiotf.APIException(500, "Devices don't support the requested operation", r.json())
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def deleteDeviceManagementRequest(self, requestId):
        """
        Clears the status of a device management request.
        You can use this operation to clear the status for a completed request, or for an in-progress request which may never complete due to a problem.
        It accepts requestId (string) as parameters
        In case of failure it throws APIException
        """
        mgmtRequests = ApiClient.mgmtSingleRequest % (self.host, requestId)
        r = requests.delete(mgmtRequests, auth=self.credentials, verify=self.verify)

        status = r.status_code
        if status == 204:
            self.logger.debug("Request status cleared")
            return True
        #403 and 404 error code needs to be added in Swagger documentation
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "Request Id not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def getDeviceManagementRequest(self, requestId):
        """
        Gets details of a device management request.
        It accepts requestId (string) as parameters
        In case of failure it throws APIException
        """
        mgmtRequests = ApiClient.mgmtSingleRequest % (self.host, requestId)
        r = requests.get(mgmtRequests, auth=self.credentials, verify=self.verify)

        status = r.status_code

        if status == 200:
            self.logger.debug("Retrieving single management request")
            return r.json()
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "Request Id not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def getDeviceManagementRequestStatus(self, requestId):
        """
        Get a list of device management request device statuses.
        In case of failure it throws APIException
        """
        mgmtRequests = ApiClient.mgmtRequestStatus % (self.host, requestId)
        r = requests.get(mgmtRequests, auth=self.credentials, verify=self.verify)

        status = r.status_code

        if status == 200:
            self.logger.debug("Retrieved all device management request statuses")
            return r.json()
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "Request status not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    def getDeviceManagementRequestStatusByDevice(self, requestId, typeId, deviceId):
        """
        Get an individual device mangaement request device status.
        In case of failure it throws APIException
        """
        mgmtRequests = ApiClient.mgmtRequestSingleDeviceStatus % (self.host, requestId, typeId, deviceId)
        r = requests.get(mgmtRequests, auth=self.credentials, verify=self.verify)

        status = r.status_code

        if status == 200:
            self.logger.debug("Retrieved device management request status of single device")
            return r.json()
        elif status == 403:
            raise ibmiotf.APIException(403, "The authentication method is invalid or the api key used does not exist", None)
        elif status == 404:
            raise ibmiotf.APIException(404, "Request status not found", None)
        elif status == 500:
            raise ibmiotf.APIException(500, "Unexpected error", None)
        else:
            raise ibmiotf.APIException(None, "Unexpected error", None)


    """
    ===========================================================================
    Device Management Extension API
        - List all device management extension packages
        - Create a new device management extension package
        - Delete a device management extension package
        - Get a specific device management extension package
        - Update a device management extension package
    ===========================================================================
    """

    def getAllDeviceManagementExtensionPkgs(self):
        """
        List all device management extension packages
        """
        dmeReq = ApiClient.dmeRequests % (self.host)
        r = requests.get(dmeReq, auth=self.credentials, verify=self.verify)
        status = r.status_code
        if status == 200:
            self.logger.debug("Retrieved all Device Management Extension Packages")
            return r.json()
        else:
            raise ibmiotf.APIException(status,"HTTP Error in getAllDeviceManagementExtensionPkgs", r)

    def createDeviceManagementExtensionPkg(self, dmeData):
        """
        Create a new device management extension package
        In case of failure it throws APIException
        """
        dmeReq = ApiClient.dmeRequests % (self.host)
        r = requests.post(dmeReq, auth=self.credentials, data=json.dumps(dmeData),
                      headers = {'content-type': 'application/json'}, verify=self.verify)
        status = r.status_code
        if status == 201:
            self.logger.debug("The DME package request has been accepted for processing")
            return r.json()
        else:
            raise ibmiotf.APIException(status,"HTTP Error in createDeviceManagementExtensionPkg", r)

    def deleteDeviceManagementExtensionPkg(self, bundleId):
        """
        Delete a device management extension package
        It accepts bundleId (string) as parameters
        In case of failure it throws APIException
        """
        dmeReq = ApiClient.dmeSingleRequest % (self.host, bundleId)
        r = requests.delete(dmeReq, auth=self.credentials, verify=self.verify)
        status = r.status_code
        if status == 204:
            self.logger.debug("Device Management Extension Package removed")
            return True
        else:
            raise ibmiotf.APIException(status,"HTTP Error in deleteDeviceManagementExtensionPkg", r)

    def getDeviceManagementExtensionPkg(self, bundleId):
        """
        Get a specific device management extension package
        It accepts bundleId (string) as parameters
        In case of failure it throws APIException
        """
        dmeReq = ApiClient.dmeSingleRequest % (self.host, bundleId)
        r = requests.get(dmeReq, auth=self.credentials, verify=self.verify)
        status = r.status_code
        if status == 200:
            self.logger.debug("Device Management Extension Package retrieved")
            return r.json()
        else:
            raise ibmiotf.APIException(status,"HTTP Error in getDeviceManagementExtensionPkg", r)

    def updateDeviceManagementExtensionPkg(self, bundleId, dmeData):
        """
        Update a device management extension package
        It accepts bundleId (string) as parameters
        In case of failure it throws APIException
        """
        dmeReq = ApiClient.dmeSingleRequest % (self.host, bundleId)
        r = requests.put(dmeReq, auth=self.credentials, data=json.dumps(dmeData),
                      headers = {'content-type': 'application/json'}, verify=self.verify)
        status = r.status_code
        if status == 200:
            self.logger.debug("Device Management Extension Package updated")
            return r.json()
        else:
            raise ibmiotf.APIException(status,"HTTP Error in updateDeviceManagementExtensionPkg", r)

    """
    ===========================================================================
    Information Management Schema APIs
    ===========================================================================
    """

    def getSchemas(self, draft=False, name=None, schemaType=None):
        """
        Get all schemas for the org.  In case of failure it throws APIException
        """
        if draft:
            req = ApiClient.allSchemasUrl % (self.host, "/draft")
        else:
            req = ApiClient.allSchemasUrl % (self.host, "")

        if name or schemaType:
            req += "?"
            if name:
                req += "name="+name
            if schemaType:
                if name:
                    req += "&"
                req += "schemaType="+schemaType

        resp = requests.get(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("All schemas retrieved")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error getting all schemas", resp)
        return [x["id"] for x in resp.json()["results"]], resp.json()

    def getSchema(self, schemaId, draft=False):
        """
        Get a single schema.  Throws APIException on failure
        """
        if draft:
            req = ApiClient.oneSchemaUrl % (self.host, "/draft", schemaId)
        else:
            req = ApiClient.oneSchemaUrl % (self.host, "", schemaId)

        resp = requests.get(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("One schema retrieved")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error in get a schema", resp)
        return resp.json()

    def createSchema(self, schemaName, schemaFileName, schemaContents, description=None):
        """
        Create a schema for the org.
        Returns: schemaId (string), response (object).
        Throws APIException on failure
        """
        req = ApiClient.allSchemasUrl % (self.host, "/draft")
        fields={
        'schemaFile': (schemaFileName, schemaContents, 'application/json'),
            'schemaType': 'json-schema',
            'name': schemaName,
        }
        if description:
            fields["description"] = description

        multipart_data = MultipartEncoder(fields=fields)
        resp = requests.post(req, auth=self.credentials, data=multipart_data,
                            headers={'Content-Type': multipart_data.content_type}, verify=self.verify)
        if resp.status_code == 201:
            self.logger.debug("Schema created")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error creating a schema", resp)
        return resp.json()["id"], resp.json()

    def deleteSchema(self, schemaId):
        """
        Delete a schema.  Parameter: schemaId (string). Throws APIException on failure.
        """
        req = ApiClient.oneSchemaUrl % (self.host, "/draft", schemaId)
        resp = requests.delete(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 204:
            self.logger.debug("Schema deleted")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error deleting schema", resp)
        return resp

    def updateSchema(self, schemaId, schemaDefinition):
        """
        Update a schema. Throws APIException on failure.
        """
        req = ApiClient.oneSchemaUrl % (self.host, "/draft", schemaId)
        body = {"schemaDefinition": schemaDefinition}
        resp = requests.put(req, auth=self.credentials, headers={"Content-Type":"application/json"},
                           data=json.dumps(body),  verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("Schema updated")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error updating schema", resp)
        return resp.json()

    def getSchemaContent(self, schemaId, draft=False):
        """
        Get the content for a schema.  Parameters: schemaId (string), draft (boolean). Throws APIException on failure.
        """
        if draft:
            req = ApiClient.oneSchemaContentUrl % (self.host, "/draft", schemaId)
        else:
            req = ApiClient.oneSchemaContentUrl % (self.host, "", schemaId)
        resp = requests.get(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("Schema content retrieved")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error getting schema content", resp)
        return resp.json()

    def updateSchemaContent(self, schemaId, schemaFile):
        """
        Updates the content for a schema.  Parameters: schemaId (string). Throws APIException on failure.
        """
        req = ApiClient.oneSchemaContentUrl % (self.host, "/draft", schemaId)
        body = {"schemaFile": schemaFile}
        resp = requests.put(req, auth=self.credentials, headers={"Content-Type":"application/json"},
                           data=json.dumps(body),  verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("Schema content updated")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error updating schema content", resp)
        return resp.json()

    """
    ===========================================================================
    Information Management event type APIs
    ===========================================================================
    """

    def getEventTypes(self, draft=False, name=None, schemaId=None):
        """
        Get all event types for an org.  Throws APIException on failure.
        """
        if draft:
            req = ApiClient.allEventTypesUrl % (self.host, "/draft")
        else:
            req = ApiClient.allEventTypesUrl % (self.host, "")

        if name or schemaId:
            req += "?"
            if name:
                req += "name="+name
            if schemaId:
                if name:
                    req += "&"
                req += "schemaId="+schemaId

        resp = requests.get(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("All event types retrieved")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error getting all event types", resp)
        return [x["id"] for x in resp.json()["results"]], resp.json()

    def createEventType(self, name, schemaId, description=None):
        """
        Creates an event type.
        Parameters: name (string), schemaId (string), description (string, optional).
        Returns: event type id (string), response (object).
        Throws APIException on failure.
        """
        req = ApiClient.allEventTypesUrl % (self.host, "/draft")
        body = {"name" : name, "schemaId" : schemaId}
        if description:
            body["description"] = description
        resp = requests.post(req, auth=self.credentials, headers={"Content-Type":"application/json"},
                            data=json.dumps(body),  verify=self.verify)
        if resp.status_code == 201:
            self.logger.debug("event type created")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error creating event type", resp)
        return resp.json()["id"], resp.json()

    def updateEventType(self, eventTypeId, name, schemaId, description=None):
        """
        Updates an event type.
        Parameters: eventTypeId (string), name (string), schemaId (string), description (string, optional).
        Throws APIException on failure.
        """
        req = ApiClient.oneEventTypesUrl % (self.host, "/draft", eventTypeId)
        body = {"name" : name, "schemaId" : schemaId}
        if description:
            body["description"] = description
        resp = requests.put(req, auth=self.credentials, headers={"Content-Type":"application/json"},
                            data=json.dumps(body),  verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("event type updated")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error updating event type", resp)
        return resp.json()

    def deleteEventType(self, eventTypeId):
        """
        Deletes an event type.  Parameters: eventTypeId (string). Throws APIException on failure.
        """
        req = ApiClient.oneEventTypeUrl % (self.host, "/draft", eventTypeId)
        resp = requests.delete(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 204:
            self.logger.debug("event type deleted")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error deleting an event type", resp)
        return resp

    def getEventType(self, eventTypeId, draft=False):
        """
        Gets an event type.  Parameters: eventTypeId (string), draft (boolean).  Throws APIException on failure.
        """
        if draft:
            req = ApiClient.oneEventTypeUrl % (self.host, "/draft", eventTypeId)
        else:
            req = ApiClient.oneEventTypeUrl % (self.host, "", eventTypeId)
        resp = requests.get(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("event type retrieved")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error getting an event type", resp)
        return resp.json()

    """
    ===========================================================================
    Information Management Physical Interface APIs
    ===========================================================================
    """

    def getPhysicalInterfaces(self, draft=False, name=None):
        """
        Get all physical interfaces for an org.
        Parameters: draft (boolean).
        Throws APIException on failure.
        """
        if draft:
            req = ApiClient.allPhysicalInterfacesUrl % (self.host, "/draft")
        else:
            req = ApiClient.allPhysicalInterfacesUrl % (self.host, "")

        if name:
            req += "?name="+name

        resp = requests.get(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("All physical interfaces retrieved")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error getting all physical interfaces", resp)
        return [x["id"] for x in resp.json()["results"]], resp.json()

    def createPhysicalInterface(self, name, description=None):
        """
        Create a physical interface.
        Parameters:
          - name (string)
          - description (string, optional)
        Returns: physical interface id, response.
        Throws APIException on failure.
        """
        req = ApiClient.allPhysicalInterfacesUrl % (self.host, "/draft")
        body = {"name" : name}
        if description:
            body["description"] = description
        resp = requests.post(req, auth=self.credentials, headers={"Content-Type":"application/json"},
                            data=json.dumps(body),  verify=self.verify)
        if resp.status_code == 201:
            self.logger.debug("physical interface created")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error creating physical interface", resp)
        return resp.json()["id"], resp.json()

    def updatePhysicalInterface(self, physicalInterfaceId, name, schemaId, description=None):
        """
        Update a physical interface.
        Parameters:
          - physicalInterfaceId (string)
          - name (string)
          - schemaId (string)
          - description (string, optional)
        Throws APIException on failure.
        """
        req = ApiClient.onePhysicalInterfacesUrl % (self.host, "/draft", physicalInterfaceId)
        body = {"name" : name, "schemaId" : schemaId}
        if description:
            body["description"] = description
        resp = requests.put(req, auth=self.credentials, headers={"Content-Type":"application/json"},
                            data=json.dumps(body),  verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("physical interface updated")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error updating physical interface", resp)
        return resp.json()

    def deletePhysicalInterface(self, physicalInterfaceId):
        """
        Delete a physical interface.
        Parameters: physicalInterfaceId (string).
        Throws APIException on failure.
        """
        req = ApiClient.onePhysicalInterfaceUrl % (self.host, "/draft", physicalInterfaceId)
        resp = requests.delete(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 204:
            self.logger.debug("physical interface deleted")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error deleting a physical interface", resp)
        return resp

    def getPhysicalInterface(self, physicalInterfaceId, draft=False):
        """
        Get a physical interface.
        Parameters:
          - physicalInterfaceId (string)
          - draft (boolean)
        Throws APIException on failure.
        """
        if draft:
            req = ApiClient.onePhysicalInterfaceUrl % (self.host, "/draft", physicalInterfaceId)
        else:
            req = ApiClient.onePhysicalInterfaceUrl % (self.host, "", physicalInterfaceId)

        resp = requests.get(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("physical interface retrieved")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error getting a physical interface", resp)
        return resp.json()


    """
    ===========================================================================
    Information Management Event Mapping APIs
    ===========================================================================
    """

    def getEvents(self, physicalInterfaceId, draft=False):
        """
        Get the event mappings for a physical interface.
        Parameters:
          - physicalInterfaceId (string)
          - draft (boolean)
        Throws APIException on failure.
        """
        if draft:
            req = ApiClient.allEventsUrl % (self.host, "/draft", physicalInterfaceId)
        else:
            req = ApiClient.allEventsUrl % (self.host, "", physicalInterfaceId)

        resp = requests.get(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("All event mappings retrieved")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error getting event mappings", resp)
        return resp.json()

    def createEvent(self, physicalInterfaceId, eventTypeId, eventId):
        """
        Create an event mapping for a physical interface.
        Parameters:
          physicalInterfaceId (string) - value returned by the platform when creating the physical interface
          eventTypeId (string) - value returned by the platform when creating the event type
          eventId (string) - matches the event id used by the device in the MQTT topic
        Throws APIException on failure.
        """
        req = ApiClient.allEventsUrl % (self.host, "/draft", physicalInterfaceId)
        body = {"eventId" : eventId, "eventTypeId" : eventTypeId}
        resp = requests.post(req, auth=self.credentials, headers={"Content-Type":"application/json"}, data=json.dumps(body),
                       verify=self.verify)
        if resp.status_code == 201:
            self.logger.debug("Event mapping created")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error creating event mapping", resp)
        return resp.json()

    def deleteEvent(self, physicalInterfaceId, eventId):
        """
        Delete an event mapping from a physical interface.
        Parameters: physicalInterfaceId (string), eventId (string).
        Throws APIException on failure.
        """
        req = ApiClient.oneEventUrl % (self.host, "/draft", physicalInterfaceId, eventId)
        resp = requests.delete(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 204:
            self.logger.debug("Event mapping deleted")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error deleting event mapping", resp)
        return resp


    """
    ===========================================================================
    Information Management Logical Interface APIs
    ===========================================================================
    """

    def getLogicalInterfaces(self, draft=False, name=None, schemaId=None):
        """
        Get all logical interfaces for an org.
        Parameters: draft (boolean).
        Returns:
            - list of ids
            - response object
        Throws APIException on failure.
        """
        if draft:
            req = ApiClient.allLogicalInterfacesUrl % (self.host, "/draft")
        else:
            req = ApiClient.allLogicalInterfacesUrl % (self.host, "")

        if name or schemaId:
            req += "?"
            if name:
                req += "name="+name
            if schemaId:
                if name:
                    req += "&"
                req += "schemaId="+schemaId

        resp = requests.get(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("All logical interfaces retrieved")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error getting all logical interfaces", resp)
        return [x["id"] for x in resp.json()["results"]], resp.json()

    def createLogicalInterface(self, name, schemaId, description=None):
        """
        Creates an logical interface..
        Parameters: name (string), schemaId (string), description (string, optional).
        Returns: logical interface id (string), response (object).
        Throws APIException on failure.
        """
        req = ApiClient.allLogicalInterfacesUrl % (self.host, "/draft")
        if description == None:
            description = ""
        body = {"name" : name, "description" : description, "schemaId" : schemaId}
        # if description:
        #   body["description"] = description
        resp = requests.post(req, auth=self.credentials, headers={"Content-Type":"application/json"},
                            data=json.dumps(body), verify=self.verify)
        if resp.status_code == 201:
            self.logger.debug("Logical interface created")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error creating logical interface", resp)
        return resp.json()["id"], resp.json()

    def updateLogicalInterface(self, logicalInterfaceId, name, schemaId, description=None):
        """
        Updates an logical interface.
        Parameters: logicalInterfaceId (string), name (string), schemaId (string), description (string, optional).
        Throws APIException on failure.
        """
        req = ApiClient.oneLogicalInterfaceUrl % (self.host, "/draft", logicalInterfaceId)
        body = {"name" : name, "schemaId" : schemaId}
        if description:
            body["description"] = description
        resp = requests.put(req, auth=self.credentials, headers={"Content-Type":"application/json"},
                            data=json.dumps(body),  verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("Logical interface updated")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error updating logical interface", resp)
        return resp.json()

    def deleteLogicalInterface(self, logicalInterfaceId):
        """
        Deletes an logical interface.
        Parameters: logicalInterfaceId (string).
        Throws APIException on failure.
        """
        req = ApiClient.oneLogicalInterfaceUrl % (self.host, "/draft", logicalInterfaceId)
        resp = requests.delete(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 204:
            self.logger.debug("logical interface deleted")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error deleting an logical interface", resp)
        return resp

    def getLogicalInterface(self, logicalInterfaceId, draft=False):
        """
        Gets an logical interface.
        Parameters:
          - logicalInterfaceId (string)
          - draft (boolean)
        Throws APIException on failure.
        """
        if draft:
            req = ApiClient.oneLogicalInterfaceUrl % (self.host, "/draft", logicalInterfaceId)
        else:
            req = ApiClient.oneLogicalInterfaceUrl % (self.host, "", logicalInterfaceId)
        resp = requests.get(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("logical interface retrieved")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error getting an logical interface", resp)
        return resp.json()

    """
    ===========================================================================
    Information Management Device Type APIs
    ===========================================================================
    """

    def addPhysicalInterfaceToDeviceType(self, typeId, physicalInterfaceId):
        """
        Adds a physical interface to a device type.
        Parameters:
            - typeId (string) - the device type
            - logicalInterfaceId (string) - the id returned by the platform on creation of the physical interface
        Throws APIException on failure.
        """
        req = ApiClient.oneDeviceTypePhysicalInterfaceUrl % (self.host, typeId)
        body = {"id" : physicalInterfaceId}
        resp = requests.post(req, auth=self.credentials, headers={"Content-Type":"application/json"}, data=json.dumps(body),
                       verify=self.verify)
        if resp.status_code == 201:
            self.logger.debug("Physical interface added to a device type")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error adding physical interface to a device type", resp)
        return resp.json()

    def removePhysicalInterfaceFromDeviceType(self, typeId):
        """
        Removes the physical interface from a device type.  Only one can be associated with a device type,
          so the physical interface id is not necessary as a parameter.
        Parameters:
                    - typeId (string) - the device type
        Throws APIException on failure.
        """
        req = ApiClient.oneDeviceTypePhysicalInterfaceUrl % (self.host, typeId)
        body = {}
        resp = requests.delete(req, auth=self.credentials, headers={"Content-Type":"application/json"}, data=json.dumps(body),
           verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("Physical interface removed from a device type")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error removing a physical interface from a device type", resp)
        return resp.json()

    def getLogicalInterfacesOnDeviceType(self, typeId, draft=False):
        """
        Get all logical interfaces for a device type.
        Parameters:
          - typeId (string)
          - draft (boolean)
        Returns:
            - list of logical interface ids
            - HTTP response object
        Throws APIException on failure.
        """
        if draft:
            req = ApiClient.allDeviceTypeLogicalInterfacesUrl % (self.host, "/draft", typeId)
        else:
            req = ApiClient.allDeviceTypeLogicalInterfacesUrl % (self.host, "", typeId)
        resp = requests.get(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("All device type logical interfaces retrieved")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error getting all device type logical interfaces", resp)
        return [appintf["id"] for appintf in resp.json()], resp.json()

    def addLogicalInterfaceToDeviceType(self, typeId, logicalInterfaceId):
        """
        Adds an logical interface to a device type.
        Parameters:
            - typeId (string) - the device type
            - logicalInterfaceId (string) - the id returned by the platform on creation of the logical interface
            - description (string) - optional (not used)
        Throws APIException on failure.
        """
        req = ApiClient.allDeviceTypeLogicalInterfacesUrl % (self.host, "/draft", typeId)
        body = {"id" : logicalInterfaceId}
#       body = {"name" : "required but not used!!!", "id" : logicalInterfaceId, "schemaId" : schemaId}
#       if description:
#           body["description"] = description
        resp = requests.post(req, auth=self.credentials, headers={"Content-Type":"application/json"}, data=json.dumps(body),
                        verify=self.verify)
        if resp.status_code == 201:
            self.logger.debug("Logical interface added to a device type")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error adding logical interface to a device type", resp)
        return resp.json()

    def removeLogicalInterfaceFromDeviceType(self, typeId, logicalInterfaceId):
        """
        Removes an logical interface from a device type.
        Parameters:
            - typeId (string) - the device type
            - logicalInterfaceId (string) - the id returned by the platform on creation of the logical interface
        Throws APIException on failure.
        """
        req = ApiClient.oneDeviceTypeLogicalInterfaceUrl % (self.host, typeId, logicalInterfaceId)
        resp = requests.delete(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 204:
            self.logger.debug("Logical interface removed from a device type")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error removing logical interface from a device type", resp)
        return resp

    def getMappingsOnDeviceType(self, typeId, draft=False):
        """
        Get all the mappings for a device type.
        Parameters:
            - typeId (string) - the device type
            - draft (boolean) - draft or active
        Throws APIException on failure.
        """
        if draft:
            req = ApiClient.allDeviceTypeMappingsUrl % (self.host, "/draft", typeId)
        else:
            req = ApiClient.allDeviceTypeMappingsUrl % (self.host, "", typeId)

        resp = requests.get(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("All device type mappings retrieved")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error getting all device type mappings", resp)
        return resp.json()

    def addMappingsToDeviceType(self, typeId, logicalInterfaceId, mappingsObject, notificationStrategy = "never"):
        """
        Add mappings for a device type.
        Parameters:
            - typeId (string) - the device type
            - logicalinterface (string) - the id of the application interface these mappings are for
            - notificationStrategy (string) - the notification strategy to use for these mappings
            - mappingsObject (Python dictionary corresponding to JSON object) example:

            { # eventid -> { property -> eventid property expression }
         "status" :  {
                "eventCount" : "($state.eventCount == -1) ? $event.d.count : ($state.eventCount+1)",
           }
      }

        Throws APIException on failure.
        """
        req = ApiClient.allDeviceTypeMappingsUrl % (self.host, "/draft", typeId)
        try:
            mappings = json.dumps({
                "logicalInterfaceId" : logicalInterfaceId,
                "notificationStrategy" : notificationStrategy,
                "propertyMappings" : mappingsObject
            })
        except Exception as exc:
            raise ibmiotf.APIException(-1, "Exception formatting mappings object to JSON", exc)
        resp = requests.post(req, auth=self.credentials, headers={"Content-Type":"application/json"}, data=mappings,
               verify=self.verify)
        if resp.status_code == 201:
            self.logger.debug("Device type mappings created for logical interface")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error creating device type mappings for logical interface", resp)
        return resp.json()

    def deleteMappingsFromDeviceType(self, typeId, logicalInterfaceId):
        """
        Deletes mappings for an application interface from a device type.
        Parameters:
            - typeId (string) - the device type
          - logicalInterfaceId (string) - the platform returned id of the application interface
        Throws APIException on failure.
        """
        req = ApiClient.oneDeviceTypeMappingUrl % (self.host, "/draft", typeId, logicalInterfaceId)
        resp = requests.delete(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 204:
            self.logger.debug("Mappings deleted from the device type")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error deleting mappings for an logical interface from a device type", resp)
        return resp

    def getMappingsOnDeviceTypeForLogicalInterface(self, typeId, logicalInterfaceId, draft=False):
        """
        Gets the mappings for an logical interface from a device type.
        Parameters:
            - typeId (string) - the device type
            - logicalInterfaceId (string) - the platform returned id of the logical interface
        Throws APIException on failure.
        """
        if draft:
            req = ApiClient.oneDeviceTypeMappingUrl % (self.host, "/draft", typeId, logicalInterfaceId)
        else:
            req = ApiClient.oneDeviceTypeMappingUrl % (self.host, "", typeId, logicalInterfaceId)

        resp = requests.get(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("Mappings retrieved from the device type")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error getting mappings for an logical interface from a device type", resp)
        return resp.json()



    def updateMappingsOnDeviceType(self, typeId, logicalInterfaceId, mappingsObject, notificationStrategy = "never"):
        """
        Add mappings for a device type.
        Parameters:
            - typeId (string) - the device type
            - logicalInterfaceId (string) - the id of the application interface these mappings are for
            - notificationStrategy (string) - the notification strategy to use for these mappings
            - mappingsObject (Python dictionary corresponding to JSON object) example:

            { # eventid -> { property -> eventid property expression }
         "status" :  {
                "eventCount" : "($state.eventCount == -1) ? $event.d.count : ($state.eventCount+1)",
           }
      }

        Throws APIException on failure.
        """
        req = ApiClient.oneDeviceTypeMappingUrl % (self.host, "/draft", typeId, logicalInterfaceId)
        try:
            mappings = json.dumps({
                "logicalInterfaceId" : logicalInterfaceId,
                "notificationStrategy" : notificationStrategy,
                "propertyMappings" : mappingsObject
            })
        except Exception as exc:
            raise ibmiotf.APIException(-1, "Exception formatting mappings object to JSON", exc)
        resp = requests.put(req, auth=self.credentials, headers={"Content-Type":"application/json"}, data=mappings,
               verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("Device type mappings updated for logical interface")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error updating device type mappings for logical interface", resp)
        return resp.json()

    """
    ===========================================================================
    Information Management Device APIs
    ===========================================================================
    # """

    def validateDeviceTypeConfiguration(self, typeId):
        """
        Validate the device type configuration.
        Parameters:
            - typeId (string) - the platform device type
        Throws APIException on failure.
        """
        req = ApiClient.draftDeviceTypeUrl % (self.host, typeId)
        body = {"operation" : "validate-configuration"}
        resp = requests.patch(req, auth=self.credentials, headers={"Content-Type":"application/json"}, data=json.dumps(body),
               verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("Validation for device type configuration succeeded")
        else:
            raise ibmiotf.APIException(resp.status_code, "Validation for device type configuration failed", resp)
        return resp.json()

    def activateDeviceTypeConfiguration(self, typeId):
        """
        Activate the device type configuration.
        Parameters:
            - typeId (string) - the platform device type
        Throws APIException on failure.
        """
        req = ApiClient.draftDeviceTypeUrl % (self.host, typeId)
        body = {"operation" : "activate-configuration"}
        resp = requests.patch(req, auth=self.credentials, headers={"Content-Type":"application/json"}, data=json.dumps(body),
               verify=self.verify)
        if (resp.status_code == 202):
            self.logger.debug("Activation for device type configuration succeeded")
        else:
            raise ibmiotf.APIException(resp.status_code, "Activation for device type configuration failed", resp)
        return resp.json()

    def deactivateDeviceTypeConfiguration(self, typeId):
        """
        Deactivate the device type configuration.
        Parameters:
            - typeId (string) - the platform device type
        Throws APIException on failure.
        """
        req = ApiClient.deviceTypeUrl % (self.host, typeId)
        body = {"operation" : "deactivate-configuration"}
        resp = requests.patch(req, auth=self.credentials, headers={"Content-Type":"application/json"}, data=json.dumps(body),
               verify=self.verify)
        if resp.status_code == 202:
            self.logger.debug("Deactivation for device type configuration succeeded")
        else:
            raise ibmiotf.APIException(resp.status_code, "Deactivation for device type configuration failed", resp)
        return resp.json()

    def validateLogicalInterfaceConfiguration(self, logicalInterfaceId):
        """
        Validate the logical interface configuration.
        Parameters:
            - logicalInterfaceId (string)
        Throws APIException on failure.
        """
        req = ApiClient.oneLogicalInterfaceUrl % (self.host, "/draft", logicalInterfaceId)
        body = {"operation" : "validate-configuration"}
        resp = requests.patch(req, auth=self.credentials, headers={"Content-Type":"application/json"}, data=json.dumps(body),
               verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("Validation for logical interface configuration succeeded")
        else:
            raise ibmiotf.APIException(resp.status_code, "Validation for logical interface configuration failed", resp)
        return resp.json()

    def activateLogicalInterfaceConfiguration(self, logicalInterfaceId):
        """
        Activate the logical interface configuration.
        Parameters:
            - logicalInterfaceId (string)
        Throws APIException on failure.
        """
        req = ApiClient.oneLogicalInterfaceUrl % (self.host, "/draft", logicalInterfaceId)
        body = {"operation" : "activate-configuration"}
        resp = requests.patch(req, auth=self.credentials, headers={"Content-Type":"application/json"}, data=json.dumps(body),
               verify=self.verify)
        if (resp.status_code == 202):
            self.logger.debug("Activation for logical interface configuration succeeded")
        else:
            raise ibmiotf.APIException(resp.status_code, "Activation for logical interface configuration failed", resp)
        return resp.json()

    def deactivateLogicalInterfaceConfiguration(self, logicalInterfaceId):
        """
        Deactivate the logical interface configuration.
        Parameters:
            - logicalInterfaceId (string)
        Throws APIException on failure.
        """
        req = ApiClient.oneLogicalInterfaceUrl % (self.host, "/draft", logicalInterfaceId)
        body = {"operation" : "deactivate-configuration"}
        resp = requests.patch(req, auth=self.credentials, headers={"Content-Type":"application/json"}, data=json.dumps(body),
               verify=self.verify)
        if resp.status_code == 202:
            self.logger.debug("Deactivate for logical interface configuration succeeded")
        else:
            raise ibmiotf.APIException(resp.status_code, "Deactivate for logical interface configuration failed", resp)
        return resp.json()

    def getDeviceStateForLogicalInterface(self, typeId, deviceId, logicalInterfaceId):
        """
        Gets the state for an logical interface for a device.
        Parameters:
            - typeId (string) - the platform device type
            - deviceId (string) - the platform device id
            - logicalInterfaceId (string) - the platform returned id of the logical interface
        Throws APIException on failure.
        """
        req = ApiClient.deviceStateUrl % (self.host, typeId, deviceId, logicalInterfaceId)
        resp = requests.get(req, auth=self.credentials, verify=self.verify)
        if resp.status_code == 200:
            self.logger.debug("State retrieved from the device type for an logical interface")
        else:
            raise ibmiotf.APIException(resp.status_code, "HTTP error getting state for an logical interface from a device type", resp)
        return resp.json()



# *****************************************************************************
# Copyright (c) 2014, 2017 IBM Corporation and other Contributors.
#
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#
# Contributors:
#   David Parker
#   Paul Slater
#   Ben Bakowski
#   Amit M Mangalvedkar
#   Lokesh Haralakatta
# *****************************************************************************

import sys
import os
import time
import json
import socket
import ssl
import logging
from logging.handlers import RotatingFileHandler
import paho.mqtt.client as paho
import threading
import iso8601
import pytz
from datetime import datetime
from encodings.base64_codec import base64_encode

__version__ = "0.3.0"

class Message:
    def __init__(self, data, timestamp=None):
        self.data = data
        self.timestamp = timestamp

class AbstractClient:
    def __init__(self, domain, organization, clientId, username, password, port=8883, logHandlers=None, cleanSession="true"):
        self.organization = organization
        self.username = username
        self.password = password
        self.address = organization + ".messaging." + domain
        self.port = port
        self.keepAlive = 60

        self.connectEvent = threading.Event()

        self._recvLock = threading.Lock()
        self._messagesLock = threading.Lock()

        self.messages = 0
        self.recv = 0

        self.clientId = clientId

        # Configure logging
        self.logger = logging.getLogger(self.__module__+"."+self.__class__.__name__)
        self.logger.setLevel(logging.INFO)

        # Remove any existing log handlers we may have picked up from getLogger()
        self.logger.handlers = []

        if logHandlers:
            if isinstance(logHandlers, list):
                # Add all supplied log handlers
                for handler in logHandlers:
                    self.logger.addHandler(handler)
            else:
                # Add the supplied log handler
                self.logger.addHandler(logHandlers)
        else:
            # Generate a default rotating file log handler and stream handler
            logFileName = '%s.log' % (clientId.replace(":", "_"))
            fhFormatter = logging.Formatter('%(asctime)-25s %(name)-25s ' + ' %(levelname)-7s %(message)s')
            rfh = RotatingFileHandler(logFileName, mode='a', maxBytes=1024000 , backupCount=0, encoding=None, delay=True)
            rfh.setFormatter(fhFormatter)

            ch = logging.StreamHandler()
            ch.setFormatter(fhFormatter)
            ch.setLevel(logging.DEBUG)

            self.logger.addHandler(rfh)
            self.logger.addHandler(ch)

        self.client = paho.Client(self.clientId, clean_session=False if cleanSession == "false" else True)

        try:
            self.tlsVersion = ssl.PROTOCOL_TLSv1_2
        except:
            self.tlsVersion = None

        # Configure authentication
        if self.username is not None:
            # In environments where either ssl is not available, or TLSv1.2 is not available we will fallback to MQTT over TCP
            if self.tlsVersion is not None:
                # Path to certificate
                caFile = os.path.dirname(os.path.abspath(__file__)) + "/messaging.pem"
                self.client.tls_set(ca_certs=caFile, certfile=None, keyfile=None, cert_reqs=ssl.CERT_REQUIRED, tls_version=ssl.PROTOCOL_TLSv1_2)
            else:
                self.port = 1883
                self.logger.warning("Unable to encrypt messages because TLSv1.2 is unavailable (MQTT over SSL requires at least Python v2.7.9 or 3.4 and openssl v1.0.1)")
            self.client.username_pw_set(self.username, self.password)

        # Attach MQTT callbacks
        self.client.on_log = self.on_log
        self.client.on_connect = self.on_connect
        self.client.on_disconnect = self.on_disconnect
        self.client.on_publish = self.on_publish

        # Initialize default message encoders and decoders.
        self._messageEncoderModules = {}

        self.start = time.time()

        # initialize callbacks
        self._onPublishCallbacks = {}


    def getMessageEncoderModule(self, messageFormat):
        return self._messageEncoderModules[messageFormat]

    def setMessageEncoderModule(self, messageFormat, module):
        self._messageEncoderModules[messageFormat] = module

    def logAndRaiseException(self, e):
        self.logger.critical(str(e))
        raise e

    def connect(self):
        self.logger.debug("Connecting... (address = %s, port = %s, clientId = %s, username = %s, password = %s)" % (self.address, self.port, self.clientId, self.username, self.password))
        try:
            self.connectEvent.clear()
            self.client.connect(self.address, port=self.port, keepalive=self.keepAlive)
            self.client.loop_start()
            if not self.connectEvent.wait(timeout=30):
                self.client.loop_stop()
                self.logAndRaiseException(ConnectionException("Operation timed out connecting to IBM Watson IoT Platform: %s" % (self.address)))

        except socket.error as serr:
            self.client.loop_stop()
            self.logAndRaiseException(ConnectionException("Failed to connect to IBM Watson IoT Platform: %s - %s" % (self.address, str(serr))))

    def disconnect(self):
        #self.logger.info("Closing connection to the IBM Watson IoT Platform")
        self.client.disconnect()
        # If we don't call loop_stop() it appears we end up with a zombie thread which continues to process
        # network traffic, preventing any subsequent attempt to reconnect using connect()
        self.client.loop_stop()
        #self.stats()
        self.logger.info("Closed connection to the IBM Watson IoT Platform")

    def stats(self):
        elapsed = ((time.time()) - self.start)

        msgPerSecond = 0 if self.messages == 0 else elapsed/self.messages
        recvPerSecond = 0 if self.recv == 0 else elapsed/self.recv
        self.logger.debug("Messages published : %s, life: %.0fs, rate: 1/%.2fs" % (self.messages, elapsed, msgPerSecond))
        self.logger.debug("Messages received  : %s, life: %.0fs, rate: 1/%.2fs" % (self.recv, elapsed, recvPerSecond))


    def on_log(self, mqttc, obj, level, string):
        self.logger.debug("%s" % (string))



    '''
    This is called when the client disconnects from the broker. The rc parameter indicates the status of the disconnection.
    When 0 the disconnection was the result of disconnect() being called, when 1 the disconnection was unexpected.
    '''
    def on_disconnect(self, mosq, obj, rc):
        if rc == 1:
            self.logger.error("Unexpected disconnect from the IBM Watson IoT Platform")
        else:
            self.logger.info("Disconnected from the IBM Watson IoT Platform")
        self.stats()

    '''
    This is called when a message from the client has been successfully sent to the broker.
    The mid parameter gives the message id of the successfully published message.
    '''
    def on_publish(self, mosq, obj, mid):
        with self._messagesLock:
            self.messages = self.messages + 1
            if mid in self._onPublishCallbacks:
                midOnPublish = self._onPublishCallbacks.get(mid)
                del self._onPublishCallbacks[mid]
                midOnPublish()
            else:
                # record the fact that paho callback has already come through so it can be called inline
                # with the publish.
                self._onPublishCallbacks[mid] = None

    '''
    Setter and Getter methods to set and get user defined keepAlive Interval  value to
    override the MQTT default value of 60
    '''
    def setKeepAliveInterval(self, newKeepAliveInterval):
        self.keepAlive = newKeepAliveInterval

    def getKeepAliveInterval(self):
        return(self.keepAlive)



'''
Generic Connection exception "Something went wrong"
'''
class ConnectionException(Exception):
    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return self.reason

'''
Specific Connection exception where the configuration is invalid
'''
class ConfigurationException(ConnectionException):
    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return self.reason


'''
Specific Connection exception where the authentication method specified is not supported
'''
class UnsupportedAuthenticationMethod(ConnectionException):
    def __init__(self, method):
        self.method = method

    def __str__(self):
        return "Unsupported authentication method: %s" % self.method


'''
Specific exception where and Event object can not be constructed
'''
class InvalidEventException(Exception):
    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return "Invalid Event: %s" % self.reason


'''
Specific exception where and Event object can not be constructed
'''
class MissingMessageDecoderException(Exception):
    def __init__(self, format):
        self.format = format

    def __str__(self):
        return "No message decoder defined for message format: %s" % self.format


class MissingMessageEncoderException(Exception):
    def __init__(self, format):
        self.format = format

    def __str__(self):
        return "No message encoder defined for message format: %s" % self.format


'''
This exception has been added in V2 and provides the following
1) The exact HTTP Status Code
2) The error thrown
3) The JSON message returned
'''
class APIException(Exception):
    def __init__(self, httpCode, message, response):
        self.httpCode = httpCode
        self.message = message
        self.response = response

    def __str__(self):
        return "[%s] %s" % (self.httpCode, self.message)

class HttpAbstractClient:
    def __init__(self, clientId, logHandlers=None):
        # Configure logging
        self.logger = logging.getLogger(self.__module__+"."+self.__class__.__name__)
        self.logger.setLevel(logging.INFO)

        # Remove any existing log handlers we may have picked up from getLogger()
        self.logger.handlers = []

        if logHandlers:
            if isinstance(logHandlers, list):
                # Add all supplied log handlers
                for handler in logHandlers:
                    self.logger.addHandler(handler)
            else:
                # Add the supplied log handler
                self.logger.addHandler(logHandlers)
        else:
            # Generate a default rotating file log handler and stream handler
            logFileName = '%s.log' % (clientId.replace(":", "_"))
            fhFormatter = logging.Formatter('%(asctime)-25s %(name)-25s ' + ' %(levelname)-7s %(message)s')
            rfh = RotatingFileHandler(logFileName, mode='a', maxBytes=1024000 , backupCount=0, encoding=None, delay=True)
            rfh.setFormatter(fhFormatter)

            ch = logging.StreamHandler()
            ch.setFormatter(fhFormatter)
            ch.setLevel(logging.DEBUG)

            self.logger.addHandler(rfh)
            self.logger.addHandler(ch)

        # Initialize default message encoders and decoders.
        self._messageEncoderModules = {}

    def connect(self):
        # No-op with HTTP client (but makes it easy to switch between using http & mqtt clients in your code)
        pass

    def disconnect(self):
        # No-op with HTTP client (but makes it easy to switch between using http & mqtt clients in your code)
        pass
        
    def getMessageEncoderModule(self, messageFormat):
        return self._messageEncoderModules[messageFormat]

    def setMessageEncoderModule(self, messageFormat, module):
        self._messageEncoderModules[messageFormat] = module

    def logAndRaiseException(self, e):
        self.logger.critical(str(e))
        raise e

    def getContentType(self,dataFormat):
        '''
           Method to detect content type using given data format
        '''
        # Default content type is json
        contentType = "application/json"
        if dataFormat == "text":
            contentType = "text/plain; charset=utf-8"
        elif dataFormat == "xml":
            contentType = "application/xml"
        elif dataFormat == "bin":
            contentType = "application/octet-stream"
        else:
            contentType = "application/json"
        # Return derived content type
        return contentType



# *****************************************************************************
# Copyright (c) 2016 IBM Corporation and other Contributors.
#
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#
# Contributors:
#   Lokesh Haralakatta  - Initial Contribution
# *****************************************************************************

import ibmiotf.device
import ibmiotf.application
import uuid
import os
from ibmiotf import *
from nose.tools import *
from nose import SkipTest
import logging
import testUtils

class TestDevice(testUtils.AbstractTest):
    registeredDevice = None
    deviceClient = None
    managedClient = None

    DEVICE_TYPE = "test_device"
    DEVICE_ID = str(uuid.uuid4())
    
    @classmethod
    def setup_class(self):
        try: 
            deviceType = self.setupAppClient.api.getDeviceType(self.DEVICE_TYPE)
        except APIException as e:
            if e.httpCode == 404:
                deviceType = self.setupAppClient.api.addDeviceType(self.DEVICE_TYPE)
            else: 
                raise e
        
        self.registeredDevice = self.setupAppClient.api.registerDevice(self.DEVICE_TYPE, self.DEVICE_ID)
        
        self.options={
            "org": self.ORG_ID,
            "type": self.registeredDevice["typeId"],
            "id": self.registeredDevice["deviceId"],
            "auth-method": "token",
            "auth-token": self.registeredDevice["authToken"]
        }
        
        self.deviceClient = ibmiotf.device.Client(self.options)

        #Create default DeviceInfo Instance and associate with ManagedClient Instance
        deviceInfoObj = ibmiotf.device.DeviceInfo()
        deviceInfoObj.fwVersion = 0.0
        self.managedClient = ibmiotf.device.ManagedClient(self.options, deviceInfo=deviceInfoObj)    

    @classmethod
    def teardown_class(self):
        del self.deviceClient
        del self.managedClient
        self.setupAppClient.api.deleteDevice(self.DEVICE_TYPE, self.DEVICE_ID)
    
    
    @raises(Exception)
    def testMissingOptions(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.device.Client({})
        assert_equal(e.exception.msg, 'Missing required property: org')

    @raises(Exception)
    def testMissingOrg(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.device.Client({"org": None, "type": self.registeredDevice["typeId"], "id": self.registeredDevice["deviceId"],
                                   "auth-method": "token", "auth-token": self.registeredDevice["authToken"] })
        assert_equal(e.exception.msg, 'Missing required property: org')

    @raises(Exception)
    def testMissingType(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.device.Client({"org": self.ORG_ID, "type": None, "id": self.registeredDevice["deviceId"],
                                   "auth-method": "token", "auth-token": self.registeredDevice["authToken"] })
        assert_equal(e.exception.msg, 'Missing required property: type')

    @raises(Exception)
    def testMissingId(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.device.Client({"org": self.ORG_ID, "type": self.registeredDevice["typeId"], "id": None,
                                   "auth-method": "token", "auth-token": self.registeredDevice["authToken"]})
        assert_equal(e.exception.msg, 'Missing required property: id')

    @raises(Exception)
    def testMissingAuthMethod(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.device.Client({"org": self.ORG_ID, "type": self.registeredDevice["typeId"], "id": self.registeredDevice["deviceId"],
                                   "auth-method": None, "auth-token": self.registeredDevice["authToken"]})
        assert_equal(e.exception.msg, 'Missing required property: auth-method')

    @raises(Exception)
    def testMissingAuthToken(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.device.Client({"org": self.ORG_ID, "type": self.registeredDevice["typeId"], "id": self.registeredDevice["deviceId"],
                                   "auth-method": "token", "auth-token": None })
        assert_equal(e.exception.msg, 'Missing required property: auth-token')

    @raises(Exception)
    def testUnSupportedAuthMethod(self):
        with assert_raises(UnsupportedAuthenticationMethod) as e:
            ibmiotf.device.Client({"org": self.ORG_ID, "type": self.registeredDevice["typeId"], "id": self.registeredDevice["deviceId"],
                                   "auth-method": "unsupported-method", "auth-token": self.registeredDevice["authToken"]})
        assert_equal(e.exception_type,UnsupportedAuthenticationMethod)

    def testDeviceClientInstance(self):
        deviceCli = ibmiotf.device.Client({"org": self.ORG_ID, "type": self.registeredDevice["typeId"], "id": self.registeredDevice["deviceId"],
                                           "auth-method": "token", "auth-token": self.registeredDevice["authToken"]})
        assert_is_instance(deviceCli , ibmiotf.device.Client)

    @raises(Exception)
    def testMissingConfigFile(self):
        deviceFile="InvalidFile.out"
        with assert_raises(ConfigurationException) as e:
            ibmiotf.device.ParseConfigFile(deviceFile)
        assert_equal(e.exception.msg, 'Error reading device configuration file')

    @raises(Exception)
    def testInvalidConfigFile(self):
        deviceFile="nullValues.conf"
        with assert_raises(AttributeError) as e:
            ibmiotf.device.ParseConfigFile(deviceFile)
        assert_equal(e.exception, AttributeError)
    
    @SkipTest
    def testNotAuthorizedConnect(self):
        client = ibmiotf.device.Client({"org": self.ORG_ID, "type": self.registeredDevice["typeId"], "id": self.registeredDevice["deviceId"],
                                              "auth-method": "token", "auth-token": "MGhUixxxxxxxxxxxx", "auth-key":"a-xxxxxx-s1tsofmoxo"})
        with assert_raises(ConnectionException) as e:
            client.connect()
        assert_equals(e.exception, ConnectionException)
        assert_equals(e.exception.msg,'Not authorized')

    @SkipTest
    def testMissingMessageEncoder(self):
        with assert_raises(MissingMessageEncoderException)as e:
            myData={'name' : 'foo', 'cpu' : 60, 'mem' : 50}
            self.deviceClient.connect()
            self.deviceClient.publishEvent("missingMsgEncode", "jason", myData)
        assert_equals(e.exception, MissingMessageEncoderException)
    
    def testDeviceInfoInstance(self):
        deviceInfoObj = ibmiotf.device.DeviceInfo()
        assert_is_instance(deviceInfoObj, ibmiotf.device.DeviceInfo)
        print(deviceInfoObj)

    def testDeviceFirmwareInstance(self):
        deviceFWObj = ibmiotf.device.DeviceFirmware()
        assert_is_instance(deviceFWObj, ibmiotf.device.DeviceFirmware)
        print(deviceFWObj)

    def testManagedClientInstance(self):
        managedClient = ibmiotf.device.ManagedClient(self.options)
        assert_is_instance(managedClient, ibmiotf.device.ManagedClient)

    def testKeepAliveIntervalMethods(self):
        assert_equals(self.deviceClient.getKeepAliveInterval(), 60)
        self.deviceClient.setKeepAliveInterval(120)
        self.deviceClient.connect()
        assert_equals(self.deviceClient.getKeepAliveInterval(), 120)
        self.deviceClient.disconnect()
        
    def testPublishEvent(self):
        def devPublishCallback():
            print("Device Publish Event done!!!")

        myData={'name' : 'foo', 'cpu' : 60, 'mem' : 50}
        self.deviceClient.connect()
        assert_true(self.deviceClient.publishEvent("testPublishJsonEvent", "json", myData,on_publish=devPublishCallback,qos=2))
        assert_true(self.deviceClient.publishEvent("testPublishXMLEvent", "xml", myData,on_publish=devPublishCallback,qos=2))
        self.deviceClient.disconnect()
    


import sys
sys.path.insert(0, 'src')

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name='ibmiotf',
    version="0.3.0",
    author='David Parker',
    author_email='parkerda@uk.ibm.com',
    package_dir={'': 'src'},
    packages=['ibmiotf', 'ibmiotf.codecs'],
    package_data={'ibmiotf': ['*.pem']},
    url='https://github.com/ibm-watson-iot/iot-python',
    license=open('LICENSE').read(),
    description='Python Client for IBM Watson IoT Platform',
    long_description=open('README.rst').read(),
    install_requires=[
        "iso8601 >= 0.1.10",
        "paho-mqtt >= 1.2",
        "pytz >= 2014.7",
        "requests >= 2.5.0",
        "requests_toolbelt >= 0.7.0",
        "dicttoxml >= 1.7.4",
        "xmltodict >= 0.10.2"
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Communications',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)
